import argparse
import csv
import os
import logging
import json
import torch
import re
import librosa
import torchaudio
import torchmetrics

import numpy as np
import matplotlib.pyplot as plt

from sklearn.metrics import balanced_accuracy_score, f1_score, accuracy_score


from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils


def load_model_dir(dir):
    weight_files = [i for i in os.listdir(dir) if i.endswith('.pt')]
    json_files = [i for i in os.listdir(dir) if i.endswith('.json')]
    mo    with open(path_model_args, 'r') as fp:
        # with io.open(path_model_args, 'r', encoding='utf8') as fp:
        model_args = json.load(fp)
        args = vars(args)
        args.update(model_args)
        args = DictToObj(**args)
        model_args = DictToObj(**model_args)dels_to_use = []
    for json_file in json_files:
        json_filename = json_file.split('.')[0]
        for weight_file in weight_files:
            if json_filename in weight_file:
                models_to_use.append(os.path.join(dir, json_file))
                break

    return models_to_use


def model_init(path_model_args):
    global args
    
        if args.is_cuda:
            model_args.device = 'cuda'
        else:
            model_args.device = 'cpu'

    path_model = path_model_args.replace('.json', '.pt')

    Model = getattr(__import__('modules_core.' + model_args.model, fromlist=['Model']), 'Model')
    model = Model(model_args)

    if torch.cuda.is_available() and args.is_cuda:
        args.device = 'cuda'
        model = model.cuda()
        model.load_state_dict(torch.load(path_model, map_location=lambda storage, loc: storage.cuda()))
    else:
        args.device = 'cpu'
        model.load_state_dict(torch.load(path_model, map_location='cpu'))

    model = model.eval()
    logging.info('model loaded')
    return model, model_args


def metrics_calc(y, x_out, model_name):
    try:
        y_prim = torch.argmax(torch.softmax(x_out, dim=1), dim=1)
        f1 = round(f1_score(y, y_prim, average='weighted') * 100.0, 3)
        # acc = round(balanced_accuracy_score(y, y_prim) * 100.0, 3)
        acc = round(accuracy_score(y, y_prim) * 100.0, 3)
    except:
        acc = None
        f1 = None

    logging.info(f'model: {model_name}, f1: {f1}, acc: {acc}')

    return {'model': model_name, 'f1': f1, 'acc': acc}


def plot_confidence_matrix(y, x_out, model_name, out_dir):
    y_prim = torch.argmax(torch.softmax(x_out, dim=1), dim=1)
    conf_matrix = np.zeros((len(args.classes), len(args.classes)))
    for idx, y_prim_idx in enumerate(y_prim):
        y_idx = y[idx]
        conf_matrix[y_idx, y_prim_idx] += 1

    plt.clf()
    plt.figure(figsize=(6, 7))
    plt.tight_layout()
    plt.title(f'infernece conf_matrix')
    plt.imshow(conf_matrix.transpose(), interpolation='nearest', cmap=plt.get_cmap('Greys'))
    titles = np.arange(len(args.classes)).astype(int)
    for i in range(len(titles)):
        for j in range(len(titles)):
            plt.annotate(
                str(round(conf_matrix[i, j], 1)),
                xy=(i, j),
                horizontalalignment='center',
                verticalalignment='center',
                backgroundcolor='white'
            )
    plt.xticks(titles, args.classes, rotation=45)
    plt.yticks(titles, args.classes, rotation=45)
    plt.xlabel('True class')
    plt.ylabel('Predicted class')
    plt.savefig(f'{out_dir}/{model_name}-unweighted_conf_matrix.png')
    plt.close()


def process_audio(audio):
    audio = torch.tensor(audio, dtype=torch.float32)
    audios = torch.split(audio.unsqueeze(dim=0), args.input_size, dim=-1)  # split in several s
    if len(audios) > 1:
        audios = audios[:len(audios)-1]
    audios = torch.cat(audios, dim=0)
    return audios


def standardize(x, x_len):
    eps = 1e-7
    mean = torch.mean(x[:x_len])
    std = torch.std(x[:x_len])
    if std == eps:
        eps = 1e-6
    x = (x - mean) / (std + eps)
    x[x_len:] = 0
    return x


def main():
    samples = [it for it in os.listdir(args.path_data_in)]
    y = []

    audio_data = []
    y_data = []
    audios = torch.empty((0,))
    for idx, name in enumerate(samples):
        if '.wav' not in name:
            continue
        audio, _ = librosa.load(os.path.join(args.path_data_in, name), sr=16000)
        audio = torch.tensor(audio)
        if audio is None:
            break
        label = re.search("happiness|sadness|anger|neutral|other", name)
        # gender = re.search("female|male", name)[0]
        gender = 'any'
        if label[0] not in args.classes:
            continue
        elif not (args.inference_gender == 'any' or args.inference_gender == gender):
            continue
        else:
            y = args.classes.index(label[0])
            audio = process_audio(audio)
            audio_data.append(audio)
            y_data.append(y)

    scores_global = []

    for path_model_args in args.path_model_args:
        model_name = path_model_args.split('.')[0].split(os.sep)[-1]

        model, model_args = model_init(path_model_args)

        out_dir = os.path.join(args.path_data_out, model_name)
        FileUtils.createDir(out_dir)

        y_all = torch.empty((1,))
        x_out_all = torch.empty((1,))
        batch = 0
        dropped_off_samples = 0

        softmax = torch.nn.Softmax(dim=1)

        logging.info('model forward')
        for idx, audio in enumerate(audio_data):
            with torch.no_grad():
                x_out = model.forward(audio.to(args.device))
                x_out = softmax(x_out)
                if x_out.max() >= args.classification_threshold:
                    x_out = torch.zeros(x_out.shape).scatter(1, x_out.argmax(1).unsqueeze(1), 1.0)

                    y = y_data[idx]

                    if batch == 0:
                        y_all = torch.tensor(y).unsqueeze(dim=0)
                        x_out_all = x_out
                    else:
                        y_all = torch.cat((y_all, torch.tensor(y).unsqueeze(dim=0)), dim=0)
                        x_out_all = torch.cat((x_out_all, x_out), dim=0)

                    batch = batch + 1
                    # torchaudio.save(f'{out_dir}/true_{y}_pred_{torch.argmax(x_out)}_{batch}.wav', audio, 16000)
                else:
                    dropped_off_samples += 1

        # calculate metrics
        logging.info('calculate metrics')
        logging.info(f'Count of dropped off samples: {dropped_off_samples}')
        metrics = metrics_calc(y_all, x_out_all, model_name)
        plot_confidence_matrix(y_all, x_out_all, model_name, out_dir)

        scores_global.append(metrics)

    csv_path = os.path.join(args.path_data_out, 'results.csv')
    with open(csv_path, 'w', newline='') as output_file:
        dict_writer = csv.DictWriter(output_file, scores_global[0].keys())
        dict_writer.writeheader()
        dict_writer.writerows(scores_global)


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-classification_threshold', default=0.0, type=float)
    arg_parser.add_argument('-path_data_in',
                            default=r"/Users/aigaandrijanova/Downloads/emo_audio_relabelled_improv_team_for_validation_augmented",
                            type=str)
    arg_parser.add_argument('-path_data_out',
                            default=r"./inference_output",
                            type=str)
    arg_parser.add_argument('-inference_gender',
                            default=f"any",  # [male, female, any]
                            type=str)
    arg_parser.add_argument('-input_size', default=64000, type=int)
    arg_parser.add_argument('-is_cuda', default=False, type=bool)
    arg_parser.add_argument(
        '-path_model_args',
        default=r'/Users/aigaandrijanova/Downloads/Models',
        nargs='*'
    )
    arg_parser.add_argument('-classes', default=[
        'happiness',
        'anger',
        'sadness',
        'other'
    ], nargs='*')

    args, args_other = arg_parser.parse_known_args()

    rootLogger = logging.getLogger()
    logging.getLogger('numba').setLevel(logging.WARNING)  # librosa.resample() won't output logs
    logging.getLogger('matplotlib.font_manager').disabled = True
    logFormatter = logging.Formatter("%(asctime)s [%(process)d] [%(thread)d] [%(levelname)s]  %(message)s")
    rootLogger.level = logging.DEBUG  # level

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)

    logging.info('Inference script is starting')

    # parse path_model_args, can be single model or list of models
    # also supported dir and will load all available models
    if type(args.path_model_args) == list:
        if len(args.path_model_args) == 1:
            if os.path.isdir(args.path_model_args[0]):
                args.path_model_args = list(load_model_dir(args.path_model_args[0]))
    else:
        if os.path.isdir(args.path_model_args):
            args.path_model_args = list(load_model_dir(args.path_model_args))
        else:
            args.path_model_args = [args.path_model_args]

    FileUtils.deleteDir(args.path_data_out)
    FileUtils.createDir(args.path_data_out)

    main()

    logging.info('Inference script is ending')
