import argparse
import copy
import datetime
import math
import os
import time

import torch
import torchmetrics

from sklearn.cluster import SpectralClustering
from sklearn.decomposition import KernelPCA, PCA
from scipy.spatial.distance import cdist
from loguru import logger
from tqdm import tqdm
import matplotlib.pyplot as plt
from scipy.io.wavfile import write
import torchvision.utils as vutils

from modules.audio_augmentations import *
from modules.args_utils import ArgsUtils
from modules.csv_utils_2 import CsvUtils2
from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules.metric_utils import LossMeter
from modules.tensorboard_utils import TensorBoardUtils, CustomSummaryWriter


if __name__ == '__main__':
    parser = argparse.ArgumentParser(add_help=False)

    # Dir creation params
    parser.add_argument('-id', default=0, type=int)
    parser.add_argument('-sequence_name', default='sequence', type=str)
    parser.add_argument('-run_name', default='run', type=str)

    # Script imports
    # parser.add_argument('-model', default='model_emo_VQ-VAE_2D_CNN', type=str)
    parser.add_argument('-model', default='model_emo_VQ-VAE_2D_VGG', type=str)
    parser.add_argument('-datasource', default='datasource_VAE_v1', type=str)
    parser.add_argument('-is_single_cuda_device', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Datasource params
    parser.add_argument('-path_emo', default='./data/emo_audio_relabelled_threshold_2_test_train_split_other_class', type=str)
    parser.add_argument('-data_workers', default=0, type=int)
    parser.add_argument('-data_samplerate', default=16000, type=int)  # Samples are in 16k freq
    parser.add_argument('-audio_cutoff', default=100.0, type=float)

    # Spectrogram params
    parser.add_argument('-spectrogram_type', default='melspectrogram', type=str)  # [spectrogram; melspectrogram; mfcc]
    parser.add_argument('-n_fft', default=1024, type=int)
    parser.add_argument('-win_length', default=512, type=int)
    parser.add_argument('-hop_length', default=128, type=int)
    parser.add_argument('-n_mels', default=128, type=int)
    parser.add_argument('-n_mfcc', default=64, type=int)

    # Audio augmentation params
    parser.add_argument('-each_augment_proba', default=0.6, type=float)
    parser.add_argument('-is_shift_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-shift_size', default=512, type=int)

    parser.add_argument('-is_bandmask_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-bandmask_maxwidth', default=0.2, type=float)
    parser.add_argument('-bandmask_bands', default=120, type=int)

    parser.add_argument('-is_revecho_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Early stopping params
    parser.add_argument('-early_stopping_patience', default=2, type=int)
    parser.add_argument('-early_stopping_param', default='test_loss', type=str)
    parser.add_argument('-early_stopping_delta_percent', default=3e-3, type=float)
    parser.add_argument('-early_stopping_param_coef', default=-1.0, type=float)

    parser.add_argument('-dropout_rate', default=0.2, type=float)
    parser.add_argument('-encoder_hidden', default=1024, type=int)

    # VQ-VAE model params
    parser.add_argument('-num_embeddings', default=256, type=int)
    parser.add_argument('-embedding_dim', default=128, type=int)
    parser.add_argument('-beta', default=0.5, type=float)

    # Metric plot and model saving params
    parser.add_argument('-save_models', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-load_model_weights', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-path_model_weights', default='./results/vq-vae_2D_CNN_VGG_4_classes_pp-1-run-22-07-05--14-28-49.pt', type=str)
    # Main params
    parser.add_argument('-epochs', default=50, type=int)
    parser.add_argument('-batch_size', default=4, type=int)
    parser.add_argument('-test_batch_size', default=4, type=int)
    # parser.add_argument('-n_clusters', default=4, type=int)
    parser.add_argument('-learning_rate', default=1e-3, type=float)
    parser.add_argument('-augment_data', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-classes', default=[
        'happiness',
        'anger',
        'sadness',
        'other'
    ], nargs='*')
    # parser.add_argument('-exclude_classes', default=[
    #     'happiness',
    #     'other'
    # ], nargs='*')
    parser.add_argument('-include_datasets', default=[
        'ravdess',
        'emo_small',
        'cremad',
        'cmu_mosei',
        'iemocap',
        'tess',
        'meld',
        'omg'
    ], nargs='*')

    args, args_other = parser.parse_known_args()
    args = ArgsUtils.add_other_args_v2(args, args_other)

    args_dict_orig = copy.deepcopy(args.__dict__)
    args_dict = args.__dict__
    args = DictToObj(**args_dict)

    path_sequence = f'./results/{args.sequence_name}'
    args.run_name += ('-' + datetime.datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S'))
    path_run = f'./results/{args.sequence_name}/{args.run_name}'
    path_artificats = f'./artifacts/{args.sequence_name}/{args.run_name}'
    FileUtils.createDir(path_run)
    FileUtils.createDir(path_artificats)
    FileUtils.writeJSON(f'{path_run}/better_args.json', vars(args))

    CsvUtils2.create_global(path_sequence)
    CsvUtils2.create_local(path_sequence, args.run_name)

    base_name = os.path.basename(path_sequence)
    logger.add(f'{path_run}/log-{base_name}.txt')

    tensorboard_writer = CustomSummaryWriter(logdir=path_run)
    tensorboard_utils = TensorBoardUtils(tensorboard_writer)

    args.device = 'cuda'
    if not torch.cuda.is_available():
        args.device = 'cpu'
        logger.info('CUDA NOT AVAILABLE')
    else:
        logger.info('cuda devices: {}'.format(torch.cuda.device_count()))

    def get_param_count(model):
        params = list(model.parameters())
        result = 0
        for param in params:
            count_param = np.prod(param.size())  # size[0] * size[1] ...
            result += count_param
        return result

    logger.info(f'Loading model')
    Model = getattr(__import__('modules_core.' + args.model, fromlist=['Model']), 'Model')
    model = Model(args)
    logger.info(f'Model param count: {get_param_count(model)}')

    if args.load_model_weights:
        logger.info(f'Loading saved model weights.')
        model.load_state_dict(torch.load(args.path_model_weights, map_location=args.device))

    if not args.is_single_cuda_device:
        if args.device == 'cuda' and torch.cuda.device_count() > 1:
            model = torch.nn.DataParallel(
                model,
                dim=0,
                device_ids=list(range(torch.cuda.device_count())))
            logger.info(f'PARALLEL MODEL {torch.cuda.device_count()}')
    model = model.to(args.device)

    get_data_loaders = getattr(__import__('modules_core.' + args.datasource, fromlist=['get_data_loaders']),
                               'get_data_loaders')
    data_loader_train, data_loader_test = get_data_loaders(args)

    optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)

    augmentations = []
    if args.is_shift_augmentation:
        augmentations.append(Shift(proba=args.each_augment_proba, shift=args.shift_size))
    if args.is_bandmask_augmentation:
        augmentations.append(BandMask(proba=args.each_augment_proba,
                                      maxwidth=args.bandmask_maxwidth,
                                      bands=args.bandmask_bands,
                                      sample_rate=args.data_samplerate))
    if args.is_revecho_augmentation:
        augmentations.append(RevEcho(proba=args.each_augment_proba, sample_rate=args.data_samplerate))
    augmentations = torch.nn.Sequential(*augmentations)

    def forward_batch(batch, mode):
        x = batch[0].to(args.device)
        if len(augmentations) and mode == 'train':
            for idx in range(len(x)):
                x[idx] = augmentations.forward(x[idx].unsqueeze(dim=0)).squeeze(dim=0)

        embedding_loss, x_dec, x_in, perplexity, z_q = model.forward(x)
        return embedding_loss, x_dec, x_in, perplexity, z_q

    def evaluate_epoch(data_loader, mode):
        labels = []
        embeddings = []
        sample_ids = []
        with torch.no_grad():
            with tqdm(data_loader, desc=f'{args.run_name} Forward {mode} epoch: {epoch}') as t:
                for idx, batch in enumerate(t):
                    t.set_postfix(loss=metrics[f'{mode}_loss'].compute().item())
                    x = batch[0]
                    real_labels = batch[1]
                    sample_id = batch[4]
                    embedding_loss, x_dec, x_in, perplexity, z_q = model.forward(x)

                    np_real_labels = real_labels.squeeze().cpu().data.numpy()
                    np_z = z_q.cpu().data.numpy()
                    embeddings += np_z.tolist()
                    labels += np_real_labels.tolist()
                    sample_ids += list(sample_id)

            # get centers
            embeddings = np.array(embeddings)
            labels = np.array(labels)
            centers_by_classes = {}
            for l in set(labels):
                cur_l = labels == l
                cur_emb = embeddings[cur_l]
                # calculate initial center
                centers_by_classes[l] = np.mean(cur_emb, axis=0)

                # remove 90th percentile of samples
                distance_to_center = cdist(
                    cur_emb,
                    np.expand_dims(centers_by_classes[l], axis=0),
                    metric='euclidean'
                ).squeeze()
                indexes_sorted = np.argsort(distance_to_center)
                embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb) * 0.9)]]

                # recalculate centers
                centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

            # calculate accuracy
            for idx, emb in enumerate(embeddings):
                dists_to_centers = cdist(
                    np.array(list(centers_by_classes.values())),
                    np.expand_dims(emb, axis=0),
                    metric='euclidean'
                ).squeeze()
                idx_closest = np.argmin(dists_to_centers)
                y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]

                # prepare data for metrics
                y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
                label = torch.LongTensor((labels[idx],))
                metrics[f'{mode}_accuracy'].update(y_prim_idx_each, label)
                metrics[f'{mode}_f1'].update(y_prim_idx_each, label)
                metrics[f'{mode}_confusion_meter'].update(y_prim_idx_each, label)

            state[f'{mode}_accuracy'] = metrics[f'{mode}_accuracy'].compute().item()
            state[f'{mode}_f1'] = metrics[f'{mode}_f1'].compute().item()
            tensorboard_utils.addPlotConfusionMatrix(
                dataXY=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                ticks=args.classes,
                tag=f'{mode}_confusion_matrix',
                global_step=epoch,
                is_percent_added=True
            )

            if epoch % 10 == 0:
                # Save reconstructed images
                vutils.save_image(x_in,
                                  f"{path_run}/{mode}_original_spec.png",
                                  nrow=8,
                                  normalize=True)
                vutils.save_image(x_dec,
                                  f"{path_run}/{mode}_reconstructed_spec.png",
                                  nrow=8,
                                  normalize=True)

                # filter embeddings for tb
                embeddings_tb = []
                labels_tb = []
                for l in list(set(labels)):
                    cur_l = labels == l
                    embeddings_tb.append(embeddings[cur_l])
                    labels_tb.append(labels[cur_l])

                labels_tb = np.hstack(labels_tb)
                embeddings_tb = np.vstack(embeddings_tb)

                emb_proj = torch.FloatTensor(embeddings)

                # save to tensorboard projector
                tensorboard_writer.add_embedding(
                    mat=emb_proj,
                    metadata=sample_ids,
                    tag=f'{mode}_embeddings',
                    global_step=epoch
                )

                # save embedding to png
                embedding_dir = f'{path_run}/embeddings'
                FileUtils.createDir(embedding_dir)
                pca = PCA(n_components=2)
                embeddings_tb_2d = pca.fit_transform(embeddings_tb)
                plt.cla()
                plt.clf()
                scatter = plt.scatter(embeddings_tb_2d[:, -1], embeddings_tb_2d[:, -2], s=1, c=labels_tb, cmap='jet')
                plt.legend(handles=scatter.legend_elements()[0], labels=args.classes)
                plt.tight_layout(pad=0.5)
                plt.savefig(f'{embedding_dir}/{mode}_{epoch}_embeddings.png')
                plt.close()

    def forward_epoch(data_loader, mode):
        with tqdm(data_loader, desc=f'{args.run_name} Forward {mode} epoch: {epoch}') as t:
            for batch in t:
                t.set_postfix(loss=metrics[f'{mode}_loss'].compute().item())
                if mode == 'train':
                    optimizer.zero_grad()
                    model.zero_grad()

                embedding_loss, x_dec, x_in, perplexity, z_q = forward_batch(batch, mode)

                # calculate losses
                loss_rec = F.mse_loss(x_dec, x_in)
                loss = loss_rec + embedding_loss

                if mode == 'train':
                    loss.backward()
                    optimizer.step()

                metrics[f'{mode}_perplexities'].update(perplexity)
                metrics[f'{mode}_rec_loss'].update(loss_rec)
                metrics[f'{mode}_loss'].update(loss)

            state[f'{mode}_rec_loss'] = metrics[f'{mode}_rec_loss'].compute().item()
            state[f'{mode}_perplexities'] = metrics[f'{mode}_perplexities'].compute().item()
            state[f'{mode}_loss'] = metrics[f'{mode}_loss'].compute().item()

    def update_early_stopping():
        # early stopping
        percent_improvement = 0
        early_stopping = False
        if epoch > 1:
            if state_before[args.early_stopping_param] != 0:
                percent_improvement = -(state[args.early_stopping_param] - state_before[args.early_stopping_param]) / \
                                      state_before[args.early_stopping_param]
            if state[args.early_stopping_param] >= 0:
                if args.early_stopping_delta_percent > percent_improvement:
                    state['early_stopping_patience'] += 1
                else:
                    state['early_stopping_patience'] = 0
            state['early_percent_improvement'] = percent_improvement
        if state['early_stopping_patience'] >= args.early_stopping_patience or \
                math.isnan(percent_improvement):
            logger.info('early stopping')
            early_stopping = True

        return percent_improvement, early_stopping

    state = {
        'epoch': 0,
        'best_loss': -1,
        'best_accuracy': -1,
        'best_f1': -1,
        'avg_epoch_time': -1,
        'epoch_time': -1,
        'early_stopping_patience': 0,
        'early_percent_improvement': 0,
        'train_loss': -1,
        'test_loss': -1,
        'train_rec_loss': -1,
        'test_rec_loss': -1,
        'train_perplexities': -1,
        'test_perplexities': -1,
        'train_accuracy': -1,
        'test_accuracy': -1,
        'train_f1': -1,
        'test_f1': -1,
        'train_positive_distance': -1,
        'test_positive_distance': -1,
        'train_negative_distance': -1,
        'test_negative_distance': -1,
        'eval_accuracy_last': -1,
        'eval_accuracy_best': -1,
    }

    avg_time_epochs = []
    time_epoch = time.time()

    metrics = dict(
        train_loss=LossMeter(),
        test_loss=LossMeter(),
        train_perplexities=LossMeter(),
        test_perplexities=LossMeter(),
        train_rec_loss=LossMeter(),
        test_rec_loss=LossMeter(),
        train_accuracy=torchmetrics.Accuracy(),
        test_accuracy=torchmetrics.Accuracy(),
        train_f1=torchmetrics.F1Score(),
        test_f1=torchmetrics.F1Score(),
        train_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        test_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        train_positive_distance=torchmetrics.MeanMetric(),
        test_positive_distance=torchmetrics.MeanMetric(),
        train_negative_distance=torchmetrics.MeanMetric(),
        test_negative_distance=torchmetrics.MeanMetric()
    )

    best_checkpoint = {}
    for epoch in range(0, args.epochs + 1):
        state_before = copy.deepcopy(state)
        logger.info('epoch: {} / {}'.format(epoch, args.epochs))

        for key in metrics.keys():
            metrics[key].reset()

        # evaluation phase
        with torch.no_grad():
            model = model.eval()
            # test set forward
            meter_prefix = 'test'
            evaluate_epoch(data_loader_test, meter_prefix)
            forward_epoch(data_loader_test, meter_prefix)
            percent_improvement, early_stopping = update_early_stopping()
        # training phase
        # train model
        meter_prefix = 'train'
        model = model.eval()
        evaluate_epoch(data_loader_train, meter_prefix)
        model = model.train()
        forward_epoch(data_loader_train, meter_prefix)

        model_module = model
        if not args.is_single_cuda_device:
            model_module = model.module

        # save best metrics
        if epoch == 0:
            state['best_loss'] = state['test_loss']
        elif state['best_loss'] > state['test_loss']:
            state['best_loss'] = state['test_loss']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_loss.pt'))
                # save checkpoint
                best_checkpoint['model'] = model_module.state_dict()
                best_checkpoint['optimizer'] = optimizer.state_dict()
                best_checkpoint['state'] = state
                best_checkpoint['args'] = args
                torch.save(best_checkpoint, os.path.join(path_run, 'best_checkpoint.pt'))

        if epoch == 0:
            state['best_accuracy'] = state['test_accuracy']
        elif state['best_accuracy'] < state['test_accuracy']:
            state['best_accuracy'] = state['test_accuracy']
            best_state_dict = model_module.state_dict()
            if args.save_models:
                torch.save(best_state_dict, os.path.join(path_run, 'best_accuracy.pt'))

        if epoch == 0:
            state['best_f1'] = state['test_f1']
        elif state['best_f1'] < state['test_f1']:
            state['best_f1'] = state['test_f1']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_f1.pt'))

        epoch_time = (time.time() - time_epoch) / 60.0
        percent = epoch / args.epochs
        state['epoch_time'] = epoch_time

        avg_time_epochs.append(epoch_time)
        state['avg_epoch_time'] = np.average(avg_time_epochs)
        eta = ((args.epochs - epoch) * state['avg_epoch_time'])
        time_epoch = time.time()
        state['epoch'] = epoch

        # save last model
        torch.save(model_module.state_dict(), os.path.join(path_run, f'{args.run_name}.pt'))

        # log state each epoch
        tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

        CsvUtils2.add_hparams(
            path_sequence=path_sequence,
            run_name=args.run_name,
            args_dict=args.__dict__,
            metrics_dict=state,
            global_step=epoch
        )

        logger.info(
            f'{round(percent * 100, 2)}% each: {round(state["avg_epoch_time"], 2)}min, eta: {round(eta, 2)} min, train loss: {round(state["train_loss"], 5)}, '
            f'test loss: {round(state["test_loss"], 5)}, improve: {round(percent_improvement, 3)}, train acc: {round(state["train_accuracy"], 2)}, '
            f'test acc: {round(state["test_accuracy"], 2)}, train f1: {round(state["train_f1"], 2)}, test f1: {round(state["test_f1"], 2)}')

        if early_stopping:
            break

    tensorboard_writer.close()
