import argparse
import copy
import math
import datetime
import os
import time
import librosa
import re
import numpy as np
import torch
import torchmetrics
from scipy.spatial.distance import cdist
from loguru import logger
from tqdm import tqdm
from sklearn.metrics import ConfusionMatrixDisplay
from modules.audio_augmentations import *
from modules.args_utils import ArgsUtils
from modules.csv_utils_2 import CsvUtils2
from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules.metric_utils import LossMeter
from modules.tensorboard_utils import TensorBoardUtils, CustomSummaryWriter
from pytorch_metric_learning import losses
from pytorch_metric_learning.distances import LpDistance
import sklearn
from sklearn.cluster import SpectralClustering, KMeans
import matplotlib.pyplot as plt
from transformers import AutoConfig, AutoModel, Wav2Vec2FeatureExtractor

plt.rcParams["figure.figsize"] = (12,8)
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'


if __name__ == '__main__':
    parser = argparse.ArgumentParser(add_help=False)

    # Dir creation params
    parser.add_argument('-id', default=0, type=int)
    parser.add_argument('-sequence_name', default='sequence', type=str)
    parser.add_argument('-run_name', default='run', type=str)

    # Script imports
    parser.add_argument('-model', default='pretrained_audio_model', type=str)
    parser.add_argument('-model_type', default='resnet', type=str)
    parser.add_argument('-is_resnet_pretrained', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-datasource', default='datasource_contrastive_v1', type=str)
    parser.add_argument('-is_single_cuda_device', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-dropout', default=0.5, type=float)
    parser.add_argument('-z_dim', default=128, type=int)
    parser.add_argument('-gender', default='', type=str)
    parser.add_argument('-pos_margin', default=0.01, type=float)
    parser.add_argument('-neg_margin', default=1.5, type=float)

    # Datasource params
    # parser.add_argument('-path_train_emo', default=r'D:\asya_projects\emo_v4\data\emo_audio_relabelled_threshold_2_test_train_split/test_dataset.json', type=str)
    # parser.add_argument('-path_train_other', default='', type=str)
    # parser.add_argument('-path_test_emo', default=r'D:\asya_projects\emo_v4\data\emo_audio_relabelled_threshold_2_test_train_split/test_dataset.json', type=str)
    parser.add_argument('-path_train_emo', default='./inference_and_test_data/emo_audio_relabelled_threshold_3_test_train_split_v3/test_dataset.json', type=str)
    parser.add_argument('-path_train_other', default='', type=str)
    parser.add_argument('-path_test_emo', default='./inference_and_test_data/emo_audio_relabelled_threshold_3_test_train_split_v3/test_dataset.json', type=str)
    parser.add_argument('-path_validation_1', default='./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1', type=str)
    parser.add_argument('-path_validation_2', default='./inference_and_test_data/emo_audio_PP_threshold_3_and_other_from_relabelled_threshold_3_for_validation', type=str)
    parser.add_argument('-path_test_other', default='', type=str)
    parser.add_argument('-data_workers', default=8, type=int)
    parser.add_argument('-data_samplerate', default=16000, type=int)  # Samples are in 16k freq
    parser.add_argument('-input_size', default=64_000, type=int)  # used in inference

    # Audio augmentation params
    parser.add_argument('-each_augment_proba', default=0.5, type=float)
    parser.add_argument('-is_shift_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-shift_size', default=512, type=int)

    parser.add_argument('-is_bandmask_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-bandmask_maxwidth', default=0.2, type=float)
    parser.add_argument('-bandmask_bands', default=120, type=int)

    parser.add_argument('-is_revecho_augmentation', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Model params
    parser.add_argument('-resample', default=1, type=int)
    parser.add_argument('-model_mean', default=0, type=float)
    parser.add_argument('-model_std', default=0, type=float)
    parser.add_argument('-standardization', default='batch_standardization', type=str)

    # Main params
    parser.add_argument('-epochs', default=2, type=int)
    parser.add_argument('-batch_size', default=8, type=int)
    parser.add_argument('-optimizer', default='adam', type=str)
    parser.add_argument('-learning_rate', default=3e-4, type=float)

    # Metric plot and model saving params
    parser.add_argument('-save_models', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Early stopping params
    parser.add_argument('-early_stopping_patience', default=2, type=int)
    parser.add_argument('-early_stopping_param', default='test_loss', type=str)
    parser.add_argument('-early_stopping_delta_percent', default=3e-3, type=float)
    parser.add_argument('-early_stopping_param_coef', default=-1.0, type=float)

    parser.add_argument('-frozen_layers', default=True, type=lambda x: (str(x).lower() == 'true')) #freeze model except the last FC for classifiaction, else freeze only Conv layers
    parser.add_argument('-training_type', default='classification', type=str) #classification or contrastive
    parser.add_argument('-which_finetuned_model', default='russian', type=str)


    parser.add_argument('-classes', default=[
        'happiness',
        'anger',
        'sadness',
        'other'
    ], nargs='*')

    args, args_other = parser.parse_known_args()
    args = ArgsUtils.add_other_args_v2(args, args_other)

    args_dict_orig = copy.deepcopy(args.__dict__)
    args_dict = args.__dict__
    args = DictToObj(**args_dict)

    path_sequence = f'./results/{args.sequence_name}'
    args.run_name += ('-' + datetime.datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S'))
    path_run = f'./results/{args.sequence_name}/{args.run_name}'
    path_artificats = f'./artifacts/{args.sequence_name}/{args.run_name}'
    FileUtils.createDir(path_run)
    FileUtils.createDir(path_artificats)
    embedding_dir = f'{path_run}/embeddings'
    FileUtils.createDir(embedding_dir)

    if args.training_type == 'classification':
        args.z_dim = len(args.classes)

    FileUtils.writeJSON(f'{path_run}/args.json', vars(args))

    CsvUtils2.create_global(path_sequence)
    CsvUtils2.create_local(path_sequence, args.run_name)

    base_name = os.path.basename(path_sequence)
    logger.add(f'{path_run}/log-{base_name}.txt')

    tensorboard_writer = CustomSummaryWriter(logdir=path_run)
    tensorboard_utils = TensorBoardUtils(tensorboard_writer)

    args.device = 'cuda'
    if not torch.cuda.is_available():
        args.device = 'cpu'
        logger.info('CUDA NOT AVAILABLE')
    else:
        logger.info('cuda devices: {}'.format(torch.cuda.device_count()))


    if args.training_type == 'classification':
        loss_func = torch.nn.CrossEntropyLoss(reduction="mean")
    else:
        loss_func = losses.ContrastiveLoss(pos_margin=args.pos_margin, neg_margin=args.neg_margin)

    Model = getattr(__import__('modules_core.' + args.model, fromlist=['Model']), 'Model')

    logger.info(f'Loading model')
    if not args.is_single_cuda_device:
        if args.device == 'cuda' and torch.cuda.device_count() > 1:
            model = torch.nn.DataParallel(
                Model(args),
                dim=0,
                device_ids=list(range(torch.cuda.device_count()))).to(args.device)
            logger.info(f'PARALLEL MODEL {torch.cuda.device_count()}')
    else:
        model = Model(args).to(args.device)


    def get_param_count(model):
        params = list(model.parameters())
        result = 0
        for param in params:
            count_param = np.prod(param.size())  # size[0] * size[1] ...
            result += count_param
        return result

    logger.info(f'Model param count: {get_param_count(model)}')

    get_data_loaders = getattr(__import__('modules_core.' + args.datasource, fromlist=['get_data_loaders']),
                               'get_data_loaders')
    data_loader_train, data_loader_test = get_data_loaders(args)

    optimizer_func = None
    if args.optimizer == 'adam':
        optimizer_func = torch.optim.AdamW(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=args.learning_rate
        )
    elif args.optimizer == 'radam':
        optimizer_func = torch.optim.RAdam(
            model.parameters(),
            lr=args.learning_rate
        )

    augmentations = []
    if args.is_shift_augmentation:
        augmentations.append(Shift(proba=args.each_augment_proba, shift=args.shift_size))
    if args.is_bandmask_augmentation:
        augmentations.append(BandMask(proba=args.each_augment_proba,
                                      maxwidth=args.bandmask_maxwidth,
                                      bands=args.bandmask_bands,
                                      sample_rate=args.data_samplerate))
    if args.is_revecho_augmentation:
        augmentations.append(RevEcho(proba=args.each_augment_proba, sample_rate=args.data_samplerate))
    augmentations = torch.nn.Sequential(*augmentations)

    def forward_batch(batch, mode):
        x = batch[0].to(args.device)
        y = batch[1].to(args.device)

        if len(augmentations) and mode == 'train':
            for idx in range(len(x)):
                x[idx] = augmentations.forward(x[idx].unsqueeze(dim=0)).squeeze(dim=0)

        output = model.forward(x)

        return output, y

    def evaluate_epoch(data_loader, mode):
        batched_accuracy = torchmetrics.Accuracy()
        embeddings = []
        labels = []
        predicted_labels = []
        with torch.no_grad():
            with tqdm(data_loader, desc=f'{args.run_name} Evaluate {mode} epoch: {epoch}') as t:
                for batch in t:
                    z, real_labels = forward_batch(batch, mode)
                    y = real_labels.cpu()

                    if args.training_type == 'classification':
                        metrics[f'{mode}_accuracy'].update(z.cpu(), y.squeeze())
                        metrics[f'{mode}_f1'].update(z.cpu(), y.squeeze())
                        metrics[f'{mode}_confusion_meter'].update(z.cpu(), y.squeeze())

                    else:
                        output = calculate_labels_from_z(z, real_labels.squeeze())
                        batched_accuracy.update(output, y.squeeze())
                        embeddings += z.cpu().data.numpy().tolist()
                        labels += y.numpy().tolist()

            if args.training_type == 'classification':
                state[f'{mode}_accuracy'] = metrics[f'{mode}_accuracy'].compute().item()
                state[f'{mode}_f1'] = metrics[f'{mode}_f1'].compute().item()
                tensorboard_utils.addPlotConfusionMatrix(
                    dataXY=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                    ticks=args.classes,
                    tag=f'{mode}_confusion_matrix',
                    global_step=epoch,
                    is_percent_added=True
                )

                conf_matrix = ConfusionMatrixDisplay(confusion_matrix=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                                                     display_labels=args.classes)
                conf_matrix.plot(cmap=plt.cm.Blues)
                plt.savefig(f'{embedding_dir}/{mode}_{epoch}_conf_matrix_acc={round(state[f"{mode}_accuracy"], 4)}.png')
                plt.close()

            else:
                # get centers
                embeddings = np.array(embeddings)
                labels = np.array(labels).squeeze()
                centers_by_classes = {}
                for l in set(labels):
                    cur_l = labels == l
                    cur_emb = embeddings[cur_l]
                    # calculate initial center
                    centers_by_classes[l] = np.mean(cur_emb, axis=0)

                    # remove 90th percentile of samples
                    distance_to_center = cdist(
                        cur_emb,
                        np.expand_dims(centers_by_classes[l], axis=0),
                        metric='euclidean'
                    ).squeeze()
                    indexes_sorted = np.argsort(distance_to_center)
                    embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]
                    # recalculate centers
                    centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

                # calculate accuracy
                for idx, emb in enumerate(embeddings):
                    dists_to_centers = cdist(
                        np.array(list(centers_by_classes.values())),
                        np.expand_dims(emb, axis=0),
                        metric='euclidean'
                    ).squeeze()
                    idx_closest = np.argmin(dists_to_centers)
                    y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]

                    # prepare data for metrics
                    y_prim_idx_each_long = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
                    label = torch.LongTensor((labels[idx],))
                    predicted_labels.append(y_prim_idx_each)
                    metrics[f'{mode}_accuracy'].update(y_prim_idx_each_long, label)
                    metrics[f'{mode}_f1'].update(y_prim_idx_each_long, label)
                    metrics[f'{mode}_confusion_meter'].update(y_prim_idx_each_long, label)

                logger.info(f"Epoch accuracy: {metrics[f'{mode}_accuracy'].compute().item()}")
                logger.info(f"Batched accuracy: {batched_accuracy.compute().item()}")

                state[f'{mode}_accuracy'] = metrics[f'{mode}_accuracy'].compute().item()
                state[f'{mode}_f1'] = metrics[f'{mode}_f1'].compute().item()
                tensorboard_utils.addPlotConfusionMatrix(
                    dataXY=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                    ticks=args.classes,
                    tag=f'{mode}_confusion_matrix',
                    global_step=epoch,
                    is_percent_added=True
                )

                conf_matrix = ConfusionMatrixDisplay(confusion_matrix=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                                                     display_labels=args.classes)
                conf_matrix.plot(cmap=plt.cm.Blues)
                plt.savefig(f'{embedding_dir}/{mode}_{epoch}_conf_matrix_acc={round(state[f"{mode}_accuracy"], 4)}.png')
                plt.close()

                # filter embeddings for tb
                embeddings_tb = []
                labels_tb = []
                for l in list(set(labels)):
                    cur_l = labels == l
                    embeddings_tb.append(embeddings[cur_l])
                    labels_tb.append(labels[cur_l])

                labels_tb = np.hstack(labels_tb)
                embeddings_tb = np.vstack(embeddings_tb)

                # save embedding to png
                embeddings_tb_2d = pca.fit_transform(embeddings_tb)
                plt.cla()
                plt.clf()
                scatter = plt.scatter(embeddings_tb_2d[:, -1], embeddings_tb_2d[:, -2], s=1, c=labels_tb, cmap='jet')
                plt.legend(handles=scatter.legend_elements()[0], labels=data_loader.dataset.json_desc['classes'])
                plt.tight_layout(pad=0.5)
                plt.savefig(f'{embedding_dir}/{mode}_{epoch}_embeddings.png')
                plt.close()


    def calculate_labels_from_z(z, labels):
        norm = torch.norm(z.detach(), p=2, dim=1, keepdim=True)
        z = (z/norm).detach()

        #calculate centers
        centers_by_class = torch.empty(size=(len(args.classes), args.z_dim)).to(args.device)
        for idx in range(len(args.classes)):
            center = torch.mean(z[labels == idx], dim=0).detach()
            centers_by_class[idx] = center

        y_prim_idx = []
        for z_vect in z:
            dists_to_centers = torch.mean(torch.abs(centers_by_class - z_vect), dim=1)
            idx_closest = torch.argmin(dists_to_centers)
            y_prim_idx.append(idx_closest.item())

        output = torch.Tensor(y_prim_idx).to(torch.int16)
        return output

    def forward_epoch(data_loader, mode):
        with tqdm(data_loader, desc=f'{args.run_name} Forward {mode} epoch: {epoch}') as t:
            for batch in t:
                t.set_postfix(loss=metrics[f'{mode}_loss'].compute().item())
                if mode == 'train':
                    optimizer_func.zero_grad()
                    model.zero_grad()

                z, real_labels = forward_batch(batch, mode)
                real_labels = real_labels.squeeze()
                # calculate loss
                loss = loss_func.forward(z, real_labels)

                if mode == 'train':
                    loss.backward()
                    optimizer_func.step()

                loss = loss.to('cpu')
                metrics[f'{mode}_loss'](loss)

            state[f'{mode}_loss'] = metrics[f'{mode}_loss'].compute().item()

    def update_early_stopping():
        # early stopping
        percent_improvement = 0
        early_stopping = False
        if epoch > 1:
            if state_before[args.early_stopping_param] != 0:
                percent_improvement = args.early_stopping_param_coef*(state[args.early_stopping_param] - state_before[args.early_stopping_param]) / \
                                      state_before[args.early_stopping_param]
            if state[args.early_stopping_param] >= 0:
                if args.early_stopping_delta_percent > percent_improvement:
                    state['early_stopping_patience'] += 1
                else:
                    state['early_stopping_patience'] = 0
            state['early_percent_improvement'] = percent_improvement
        if state['early_stopping_patience'] >= args.early_stopping_patience or \
                math.isnan(percent_improvement):
            logger.info('early stopping')
            early_stopping = True

        return percent_improvement, early_stopping

    state = {
        'epoch': 0,
        'best_loss': -1,
        'best_accuracy': -1,
        'best_f1': -1,
        'avg_epoch_time': -1,
        'epoch_time': -1,
        'early_stopping_patience': 0,
        'early_percent_improvement': 0,
        'train_loss': -1,
        'test_loss': -1,
        'train_accuracy': -1,
        'test_accuracy': -1,
        'train_f1': -1,
        'test_f1': -1,
        'eval_accuracy_1': 0,
        'eval_accuracy_2': 0
    }

    avg_time_epochs = []
    time_epoch = time.time()

    metrics = dict(
        train_loss=LossMeter(),
        test_loss=LossMeter(),
        train_accuracy=torchmetrics.Accuracy(),
        test_accuracy=torchmetrics.Accuracy(),
        train_f1=torchmetrics.F1Score(),
        test_f1=torchmetrics.F1Score(),
        train_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        test_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
    )

    pca = sklearn.decomposition.PCA(n_components = 2)
    best_checkpoint = {}
    for epoch in range(0, args.epochs + 1):
        state_before = copy.deepcopy(state)
        logger.info('epoch: {} / {}'.format(epoch, args.epochs))

        for key in metrics.keys():
            metrics[key].reset()

        # training phase
        # train model
        meter_prefix = 'train'
        model = model.train()
        forward_epoch(data_loader_train, meter_prefix)
        model = model.eval()
        evaluate_epoch(data_loader_train, meter_prefix)
        # evaluation phase
        with torch.no_grad():
            model = model.eval()
            # test set forward
            meter_prefix = 'test'
            forward_epoch(data_loader_test, meter_prefix)
            evaluate_epoch(data_loader_test, meter_prefix)

        model_module = model
        if not args.is_single_cuda_device:
            model_module = model.module

        # save best metrics
        if epoch == 0:
            state['best_loss'] = state['test_loss']
        elif state['best_loss'] > state['test_loss']:
            state['best_loss'] = state['test_loss']
            if args.save_models:
                # torch.save(model_module.state_dict(), os.path.join(path_run, 'best_loss.pt'))
                # save checkpoint
                best_checkpoint['model'] = model_module.state_dict()
                best_checkpoint['optimizer'] = optimizer_func.state_dict()
                best_checkpoint['state'] = state
                best_checkpoint['args'] = args
                # torch.save(best_checkpoint, os.path.join(path_run, 'best_checkpoint.pt'))

        if epoch == 0:
            state['best_accuracy'] = state['test_accuracy']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_accuracy.pt'))
        elif state['best_accuracy'] < state['test_accuracy']:
            state['best_accuracy'] = state['test_accuracy']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_accuracy.pt'))

        if epoch == 0:
            state['best_f1'] = state['test_f1']
        elif state['best_f1'] < state['test_f1']:
            state['best_f1'] = state['test_f1']
            # if args.save_models:
                # torch.save(model_module.state_dict(), os.path.join(path_run, 'best_f1.pt'))

        epoch_time = (time.time() - time_epoch) / 60.0
        if args.epochs != 0:
            percent = epoch / args.epochs
        else:
            percent = 100

        state['epoch_time'] = epoch_time

        avg_time_epochs.append(epoch_time)
        state['avg_epoch_time'] = np.average(avg_time_epochs)
        eta = ((args.epochs - epoch) * state['avg_epoch_time'])
        time_epoch = time.time()
        state['epoch'] = epoch

        # save last model
        torch.save(model_module.state_dict(), os.path.join(path_run, f'{args.run_name}.pt'))

        # log state each epoch
        tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

        CsvUtils2.add_hparams(
            path_sequence=path_sequence,
            run_name=args.run_name,
            args_dict=args.__dict__,
            metrics_dict=state,
            global_step=epoch
        )

        # for best accuracy
        with torch.no_grad():
            percent_improvement, early_stopping = update_early_stopping()

        logger.info(
            f'{round(percent * 100, 2)}% each: {round(state["avg_epoch_time"], 2)}min, eta: {round(eta, 2)} min, train loss: {round(state["train_loss"], 5)}, '
            f'test loss: {round(state["test_loss"], 5)}, improve: {round(percent_improvement, 3)}, train acc: {round(state["train_accuracy"], 2)}, '
            f'test acc: {round(state["test_accuracy"], 2)}, train f1: {round(state["train_f1"], 2)}, test f1: {round(state["test_f1"], 2)}')

        if early_stopping:
            break

    # run eval
    cur_epoch = epoch
    model.load_state_dict(torch.load(os.path.join(path_run, f'best_accuracy.pt'),
                                 map_location=lambda storage, loc: storage.cuda()))
    model = model.eval()

    validation_tools = {}
    validation_tools['eval_accuracy_1'] = torchmetrics.Accuracy()
    validation_tools['eval_accuracy_2'] = torchmetrics.Accuracy()
    validation_tools['eval_confusion_meter_1'] = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
    validation_tools['eval_confusion_meter_2'] = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")

    #both validation sets
    sample_dict = {}
    validation_tools['samples_1'] = [it for it in os.listdir(args.path_validation_1)]
    validation_tools['samples_2'] = [it for it in os.listdir(args.path_validation_2)]
    validation_tools['path_validation_1'] = args.path_validation_1
    validation_tools['path_validation_2'] = args.path_validation_2
    with torch.no_grad():
        for validation_set_number in range(1, 3):
            eval_accuracy = validation_tools[f'eval_accuracy_{validation_set_number}']
            eval_confusion_meter = validation_tools[f'eval_confusion_meter_{validation_set_number}']
            samples = validation_tools[f'samples_{validation_set_number}']
            path_validation = validation_tools[f'path_validation_{validation_set_number}']
            true_labels = []
            embeddings = []
            for idx, name in enumerate(tqdm(samples, desc='Evaluate model')):
                if '.wav' not in name:
                    continue
                audio, _ = librosa.load(os.path.join(path_validation, name), sr=args.data_samplerate)
                if audio is None:
                    break
                else:
                    label = re.search("happiness|sadness|anger|neutral|other", name)[0]
                    if label not in args.classes:
                        continue

                    if args.gender and validation_set_number == 1:
                        #genders available only for improv dataset
                        gender = re.search("male|female", name)[0]
                        if gender != args.gender:
                            continue

                input_size = args.input_size
                step_size = int(args.input_size / 2)
                pad_reference = step_size
                if len(audio) < input_size:
                    pad_reference = input_size
                overflow = len(audio) % pad_reference
                if overflow > 0:
                    pad_len = pad_reference - overflow
                    audio = np.pad(audio, (0, pad_len))
                t_audio = torch.FloatTensor(audio).to(args.device).unsqueeze(dim=0).unsqueeze(dim=0).unsqueeze(dim=0)
                t_audio_strided = F.unfold(t_audio, kernel_size=(1, input_size), stride=(1, step_size)).squeeze().T

                if len(t_audio_strided.shape) == 1:
                    t_audio_strided = t_audio_strided.unsqueeze(dim=0)

                loader = torch.utils.data.DataLoader(
                    torch.utils.data.TensorDataset(t_audio_strided),
                    batch_size=args.batch_size,
                    shuffle=False
                )
                for batch, in loader:
                    if args.training_type == 'classification':
                        result = model.forward(batch).cpu()
                        label = torch.as_tensor([args.classes.index(label)])
                        eval_accuracy.update(result, label)
                        eval_confusion_meter.update(result, label)
                    else:
                        embedding = model.forward(batch).cpu()
                        for emb in embedding:
                            embeddings.append(emb)
                            true_labels.append(args.classes.index(label))

            if args.training_type != 'classification':
                embeddings = np.stack(embeddings)
                true_labels = np.array(true_labels)
                # get centers
                centers_by_classes = {}
                for l in set(true_labels):
                    cur_l = true_labels == l
                    cur_emb = embeddings[cur_l]
                    # calculate initial center
                    centers_by_classes[l] = np.mean(cur_emb, axis=0)

                    # remove 90th percentile of samples
                    distance_to_center = cdist(
                        cur_emb,
                        np.expand_dims(centers_by_classes[l], axis=0),
                        metric='euclidean'
                    ).squeeze()
                    indexes_sorted = np.argsort(distance_to_center)
                    embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

                    # recalculate centers
                    centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

                # calculate accuracy
                for idx, emb in enumerate(embeddings):
                    dists_to_centers = cdist(
                        np.array(list(centers_by_classes.values())),
                        np.expand_dims(emb, axis=0),
                        metric='euclidean'
                    ).squeeze()
                    idx_closest = np.argmin(dists_to_centers)
                    y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
                    # prepare data for metrics
                    y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
                    label = torch.LongTensor((true_labels[idx],))

                    eval_accuracy.update(y_prim_idx_each, label)
                    eval_confusion_meter.update(y_prim_idx_each, label)

                embeddings_tb_2d = pca.fit_transform(embeddings)
                plt.cla()
                plt.clf()
                scatter = plt.scatter(embeddings_tb_2d[:, -1], embeddings_tb_2d[:, -2], s=1, c=true_labels, cmap='jet')
                plt.legend(handles=scatter.legend_elements()[0], labels=args.classes)
                plt.tight_layout(pad=0.5)
                embedding_dir = f'{path_run}/embeddings'
                plt.savefig(f'{embedding_dir}/eval_embeddings_best_acc.png')
                plt.close()

            current_accuracy = eval_accuracy.compute().item()
            state[f'eval_accuracy_{validation_set_number}'] = current_accuracy
            current_accuracy = round(current_accuracy, 4)

            tensorboard_utils.addPlotConfusionMatrix(
                dataXY=eval_confusion_meter.compute().numpy(),
                ticks=args.classes,
                tag=f'eval_confusion_matrix_{validation_set_number}',
                global_step=cur_epoch,
                is_percent_added=True
            )

            conf_matrix = ConfusionMatrixDisplay(confusion_matrix=eval_confusion_meter.compute().numpy(),
                                                 display_labels=args.classes)
            conf_matrix.plot(cmap=plt.cm.Blues)
            plt.savefig(f'{embedding_dir}/eval_conf_matrix_{validation_set_number}_acc={current_accuracy}.png')
            plt.close()

            logger.info(f'Evaluate best_acc results {validation_set_number}: {current_accuracy}')

    # log eval values at end of training
    tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

    CsvUtils2.add_hparams(
        path_sequence=path_sequence,
        run_name=args.run_name,
        args_dict=args.__dict__,
        metrics_dict=state,
        global_step=epoch
    )

    tensorboard_writer.close()
