import argparse
import copy
import math
import datetime
import os
import time
import torchmetrics
import torch
import re
import librosa
import numpy as np
from sklearn.decomposition import KernelPCA, PCA
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore")

from loguru import logger
from tqdm import tqdm

from modules.audio_augmentations import *
from modules.args_utils import ArgsUtils
from modules.csv_utils_2 import CsvUtils2
from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules.metric_utils import LossMeter
from modules.tensorboard_utils import TensorBoardUtils, CustomSummaryWriter
from modules.loss_functions import OnlineTripletLoss, AAMSoftmaxLoss
from modules.proxy_nca_loss import ProxyNCALoss
from modules.triplet_utils import HardestNegativeTripletSelector, RandomNegativeTripletSelector, SemihardNegativeTripletSelector

if __name__ == '__main__':

    parser = argparse.ArgumentParser(add_help=False)

    # Dir creation params
    parser.add_argument('-id', default=0, type=int)
    parser.add_argument('-sequence_name', default='sequence', type=str)
    parser.add_argument('-run_name', default='run', type=str)
    parser.add_argument('-gender', default='', type=str)

    # Script imports
    parser.add_argument('-model', default='model_1D_demucs_triplet_no_rnn_v2', type=str)
    parser.add_argument('-datasource', default='datasource_triplet_v3', type=str)
    parser.add_argument('-is_single_cuda_device', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Datasource params
    parser.add_argument('-path_train_emo', default='./data/emo_audio_relabelled_threshold_2_test_train_split_other_class/train_dataset.json', type=str)
    parser.add_argument('-path_train_other', default='', type=str)
    parser.add_argument('-path_test_emo', default='./data/emo_audio_relabelled_threshold_2_test_train_split_other_class/test_dataset.json', type=str)
    parser.add_argument('-path_test_other', default='', type=str)
    parser.add_argument('-path_validation', default='./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1', type=str)
    parser.add_argument('-data_workers', default=0, type=int)
    parser.add_argument('-data_samplerate', default=16000, type=int)  # Samples are in 16k freq
    parser.add_argument('-data_scaling_type', default='standardization', type=str)  # [standardization; normalization; none]
    parser.add_argument('-audio_cutoff', default=100.0, type=float)
    parser.add_argument('-stratification_overhead', default=1.3, type=float)

    # Dataloader params
    parser.add_argument('-sampler_type', default='WRS', type=str)
    # ['MPCS' = MPerClassSampler; 'WRS' = WeightedRandomSampler]

    # Audio augmentation params
    parser.add_argument('-each_augment_proba', default=0.0, type=float)
    parser.add_argument('-is_shift_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-shift_size', default=512, type=int)

    parser.add_argument('-is_bandmask_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-bandmask_maxwidth', default=0.2, type=float)
    parser.add_argument('-bandmask_bands', default=120, type=int)

    parser.add_argument('-is_revecho_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))

    # Model params
    parser.add_argument('-apply_normalization', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-normalization_scope', default='sample', type=str)  # [sample; batch; dataset]
    parser.add_argument('-encoder_ch_out', default=128, type=int)
    parser.add_argument('-resample', default=1, type=int)
    parser.add_argument('-model_mean', default=0, type=float)
    parser.add_argument('-model_std', default=0, type=float)
    parser.add_argument('-normalization_module', default='instance', type=str)  # [batch; instance; none]
    parser.add_argument('-input_size', default=64_000, type=int)

    # demucs specific args
    parser.add_argument('-depth', default=5, type=int)
    parser.add_argument('-hidden_size', default=48, type=int)
    parser.add_argument('-kernel_size', default=8, type=int)
    parser.add_argument('-stride', default=4, type=int)
    # Controls if LSTM will be bidirectional (False - bidirectional, True - single direction):
    parser.add_argument('-causal', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-is_glu', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-floor', default=1e-3, type=float)
    parser.add_argument('-normalize', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-growth', default=2, type=int)
    parser.add_argument('-rescale', default=0.1, type=float)
    parser.add_argument('-max_hidden', default=10000, type=int)
    parser.add_argument('-dropout_rate', default=0.2, type=float)
    parser.add_argument('-embedding_size', default=128, type=int)

    # Spectrogram params
    parser.add_argument('-spectrogram_type', default='melspectrogram', type=str)  # [spectrogram; melspectrogram; mfcc]
    parser.add_argument('-n_fft', default=1024, type=int)
    parser.add_argument('-win_length', default=512, type=int)
    parser.add_argument('-hop_length', default=128, type=int)
    parser.add_argument('-n_mels', default=128, type=int)
    parser.add_argument('-n_mfcc', default=64, type=int)

    # Main params
    parser.add_argument('-epochs', default=4, type=int)
    parser.add_argument('-batch_size', default=27, type=int)
    parser.add_argument('-optimizer', default='adam', type=str)
    parser.add_argument('-learning_rate', default=3e-4, type=float)
    parser.add_argument('-is_quick_test', default=False, type=lambda x: (str(x).lower() == 'true'))

    # Loss params
    parser.add_argument('-loss_function', default='aam_softmax', type=str)
    parser.add_argument('-triplet_margin', default=0.2, type=float)
    parser.add_argument('-is_trainable_margin', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-aam_softmax_scale', default=30, type=float) # scale for AAMSoftmaxLoss
    parser.add_argument('-aam_softmax_margin', default=0.2, type=float) # margin for AAMSoftmaxLoss
    parser.add_argument('-embeddings_scale', default=1., type=float) # embeddings_scale for ProxyNCALoss
    parser.add_argument('-proxy_scale', default=3., type=float) # proxy_scale for ProxyNCALoss
    parser.add_argument('-smoothing_factor', default=0.1, type=float) # smoothing_factor for ProxyNCALoss


    # Metric plot and model saving params
    parser.add_argument('-save_models', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Early stopping params
    parser.add_argument('-early_stopping_patience', default=3, type=int)
    parser.add_argument('-early_stopping_param', default='test_loss', type=str)
    parser.add_argument('-early_stopping_delta_percent', default=3e-3, type=float)
    parser.add_argument('-early_stopping_param_coef', default=-1.0, type=float)

    parser.add_argument('-classes', default=[
        'happiness',
        'anger',
        'sadness',
        'other'
    ], nargs='*')

    args, args_other = parser.parse_known_args()
    args = ArgsUtils.add_other_args_v2(args, args_other)

    args_dict_orig = copy.deepcopy(args.__dict__)
    args_dict = args.__dict__
    args = DictToObj(**args_dict)

    path_sequence = f'./results/{args.sequence_name}'
    args.run_name += ('-' + datetime.datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S'))
    path_run = f'./results/{args.sequence_name}/{args.run_name}'
    path_artificats = f'./artifacts/{args.sequence_name}/{args.run_name}'
    FileUtils.createDir(path_run)
    FileUtils.createDir(path_artificats)
    FileUtils.writeJSON(f'{path_run}/args.json', vars(args))

    CsvUtils2.create_global(path_sequence)
    CsvUtils2.create_local(path_sequence, args.run_name)

    base_name = os.path.basename(path_sequence)
    logger.add(f'{path_run}/log-{base_name}.txt')

    tensorboard_writer = CustomSummaryWriter(logdir=path_run)
    tensorboard_utils = TensorBoardUtils(tensorboard_writer)

    args.device = 'cuda'
    if not torch.cuda.is_available():
        args.device = 'cpu'
        logger.info('CUDA NOT AVAILABLE')
    else:
        logger.info('cuda devices: {}'.format(torch.cuda.device_count()))

    if args.loss_function == 'triplet_hardest_neg':
        loss_func = OnlineTripletLoss(args.triplet_margin, HardestNegativeTripletSelector(args.triplet_margin), args.is_trainable_margin)
    elif args.loss_function == 'triplet_rand_neg':
        loss_func = OnlineTripletLoss(args.triplet_margin, RandomNegativeTripletSelector(args.triplet_margin), args.is_trainable_margin)
    elif args.loss_function == 'triplet_semihard_neg':
        loss_func = OnlineTripletLoss(args.triplet_margin, SemihardNegativeTripletSelector(args.triplet_margin), args.is_trainable_margin)
    elif args.loss_function == 'aam_softmax':
        loss_func = AAMSoftmaxLoss(in_features=args.embedding_size, out_features=len(args.classes), scale=args.aam_softmax_scale, margin=args.aam_softmax_margin)
    elif args.loss_function == 'proxy_nca':
        loss_func = ProxyNCALoss(n_classes=len(args.classes), embedding_size=args.embedding_size,
                                 embedding_scale=args.embeddings_scale, proxy_scale=args.proxy_scale,
                                 smoothing_factor=args.smoothing_factor)

    def get_param_count(model):
        params = list(model.parameters())
        result = 0
        for param in params:
            count_param = np.prod(param.size())  # size[0] * size[1] ...
            result += count_param
        return result

    logger.info(f'Loading model')
    Model = getattr(__import__('modules_core.' + args.model, fromlist=['Model']), 'Model')
    model = Model(args)
    logger.info(f'Model param count: {get_param_count(model)}')

    if not args.is_single_cuda_device :
        if args.device == 'cuda' and torch.cuda.device_count() > 1:
            model = torch.nn.DataParallel(
                model,
                dim=0,
                device_ids=list(range(torch.cuda.device_count())))
            logger.info(f'PARALLEL MODEL {torch.cuda.device_count()}')
    model = model.to(args.device)

    get_data_loaders = getattr(__import__('modules_core.' + args.datasource, fromlist=['get_data_loaders']),
                               'get_data_loaders')
    data_loader_train, data_loader_test = get_data_loaders(args)

    optimizer_func = None
    if args.optimizer == 'adam':
        optimizer_func = torch.optim.Adam(
            model.parameters(),
            lr=args.learning_rate
        )
    elif args.optimizer == 'radam':
        optimizer_func = torch.optim.RAdam(
            model.parameters(),
            lr=args.learning_rate
        )

    augmentations = []
    if args.is_shift_augmentation:
        augmentations.append(Shift(proba=args.each_augment_proba, shift=args.shift_size))
    if args.is_bandmask_augmentation:
        augmentations.append(BandMask(proba=args.each_augment_proba,
                                      maxwidth=args.bandmask_maxwidth,
                                      bands=args.bandmask_bands,
                                      sample_rate=args.data_samplerate))
    if args.is_revecho_augmentation:
        augmentations.append(RevEcho(proba=args.each_augment_proba, sample_rate=args.data_samplerate))
    augmentations = torch.nn.Sequential(*augmentations)

    def forward_batch(batch, mode):
        x = batch[0].to(args.device)
        y = batch[1].to(args.device)

        if len(augmentations) and mode == 'train':
            for idx in range(len(x)):
                x[idx] = augmentations.forward(x[idx].unsqueeze(dim=0)).squeeze(dim=0)

        output = model.forward(x)

        return output, y

    def evaluate_epoch(data_loader, mode):
        embeddings = []
        labels = []
        with torch.no_grad():
            with tqdm(data_loader, desc=f'{args.run_name} Evaluate {mode} epoch: {epoch}') as t:
                for batch in t:
                    output, y = forward_batch(batch, mode)

                    output_np = output.cpu().data.numpy()
                    np_y = y.squeeze().cpu().data.numpy()
                    embeddings += output_np.tolist()
                    labels += np_y.tolist()

                    if args.is_quick_test:
                        break

            # get centers
            embeddings = np.array(embeddings)
            labels = np.array(labels)
            centers_by_classes = {}
            for l in set(labels):
                cur_l = labels == l
                cur_emb = embeddings[cur_l]
                # calculate initial center
                centers_by_classes[l] = np.mean(cur_emb, axis=0)

                # remove 90th percentile of samples
                distance_to_center = cdist(
                    cur_emb,
                    np.expand_dims(centers_by_classes[l], axis=0),
                    metric='euclidean'
                ).squeeze()
                indexes_sorted = np.argsort(distance_to_center)
                embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

                # recalculate centers
                centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

            # calculate accuracy
            for idx, emb in enumerate(embeddings):
                dists_to_centers = cdist(
                    np.array(list(centers_by_classes.values())),
                    np.expand_dims(emb, axis=0),
                    metric='euclidean'
                ).squeeze()
                idx_closest = np.argmin(dists_to_centers)
                y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]

                # prepare data for metrics
                y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
                label = torch.LongTensor((labels[idx],))
                metrics[f'{mode}_accuracy'].update(y_prim_idx_each, label)
                metrics[f'{mode}_f1'].update(y_prim_idx_each, label)
                metrics[f'{mode}_confusion_meter'].update(y_prim_idx_each, label)

            logger.info(f"Epoch accuracy: {metrics[f'{mode}_accuracy'].compute().item()}")

            state[f'{mode}_accuracy'] = metrics[f'{mode}_accuracy'].compute().item()
            state[f'{mode}_f1'] = metrics[f'{mode}_f1'].compute().item()
            tensorboard_utils.addPlotConfusionMatrix(
                dataXY=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                ticks=args.classes,
                tag=f'{mode}_confusion_matrix',
                global_step=epoch,
                is_percent_added=True
            )
            # filter embeddings for tb
            embeddings_tb = []
            labels_tb = []
            for l in list(set(labels)):
                cur_l = labels == l
                embeddings_tb.append(embeddings[cur_l])
                labels_tb.append(labels[cur_l])

            labels_tb = np.hstack(labels_tb)
            embeddings_tb = np.vstack(embeddings_tb)

            # save embedding to png
            embedding_dir = f'{path_run}/embeddings'
            FileUtils.createDir(embedding_dir)
            pca = PCA(n_components=2)
            embeddings_tb_2d = pca.fit_transform(embeddings_tb)
            plt.cla()
            plt.clf()
            scatter = plt.scatter(embeddings_tb_2d[:, -1], embeddings_tb_2d[:, -2], s=1, c=labels_tb, cmap='jet')
            plt.legend(handles=scatter.legend_elements()[0], labels=data_loader.dataset.classes)
            plt.tight_layout(pad=0.5)
            plt.savefig(f'{embedding_dir}/{mode}_{epoch}_embeddings.png')
            plt.close()
        return centers_by_classes


    def forward_epoch(data_loader, mode):
        with tqdm(data_loader, desc=f'{args.run_name} Forward {mode} epoch: {epoch}') as t:
            for batch in t:
                t.set_postfix(loss=metrics[f'{mode}_loss'].compute().item())
                if mode == 'train':
                    optimizer_func.zero_grad()
                    model.zero_grad()

                output, y = forward_batch(batch, mode)

                # calculate loss
                pos_distance = 0
                neg_distance = 0
                loss = loss_func.forward(output.squeeze(), y.squeeze())
                if type(loss) is tuple:
                    loss, pos_distance, neg_distance = loss

                if mode == 'train':
                    loss.backward()
                    optimizer_func.step()

                loss = loss.to('cpu')
                metrics[f'{mode}_loss'].update(loss)
                metrics[f'{mode}_positive_distance'].update(pos_distance)
                metrics[f'{mode}_negative_distance'].update(neg_distance)

                if args.is_quick_test:
                    break

            state[f'{mode}_loss'] = metrics[f'{mode}_loss'].compute().item()
            state[f'{mode}_positive_distance'] = metrics[f'{mode}_positive_distance'].compute().item()
            state[f'{mode}_negative_distance'] = metrics[f'{mode}_negative_distance'].compute().item()

    def update_early_stopping():
        # early stopping
        percent_improvement = 0
        early_stopping = False
        if epoch > 1:
            if state_before[args.early_stopping_param] != 0:
                percent_improvement = -(state[args.early_stopping_param] - state_before[args.early_stopping_param]) / \
                                      state_before[args.early_stopping_param]
            if state[args.early_stopping_param] >= 0:
                if args.early_stopping_delta_percent > percent_improvement:
                    state['early_stopping_patience'] += 1
                else:
                    state['early_stopping_patience'] = 0
            state['early_percent_improvement'] = percent_improvement
        if state['early_stopping_patience'] >= args.early_stopping_patience or \
                math.isnan(percent_improvement):
            logger.info('early stopping')
            early_stopping = True

        return percent_improvement, early_stopping

    state = {
        'epoch': 0,
        'best_loss': -1,
        'best_accuracy': -1,
        'best_f1': -1,
        'avg_epoch_time': -1,
        'epoch_time': -1,
        'early_stopping_patience': 0,
        'early_percent_improvement': 0,
        'train_loss': -1,
        'test_loss': -1,
        'train_accuracy': -1,
        'test_accuracy': -1,
        'train_f1': -1,
        'test_f1': -1,
        'train_positive_distance': -1,
        'test_positive_distance': -1,
        'train_negative_distance': -1,
        'test_negative_distance': -1,
        'eval_accuracy_self_last': -1,
        'eval_accuracy_self_best': -1,
        'eval_accuracy_train_embs_last': -1,
        'eval_accuracy_train_embs_best': -1
    }

    avg_time_epochs = []
    time_epoch = time.time()

    metrics = dict(
        train_loss=LossMeter(),
        test_loss=LossMeter(),
        train_accuracy=torchmetrics.Accuracy(),
        test_accuracy=torchmetrics.Accuracy(),
        train_f1=torchmetrics.F1Score(),
        test_f1=torchmetrics.F1Score(),
        train_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        test_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        train_positive_distance=torchmetrics.MeanMetric(),
        test_positive_distance=torchmetrics.MeanMetric(),
        train_negative_distance=torchmetrics.MeanMetric(),
        test_negative_distance=torchmetrics.MeanMetric()
    )

    best_checkpoint = {}
    start_epoch = 1
    for epoch in range(start_epoch, args.epochs + 1):
        state_before = copy.deepcopy(state)
        logger.info('epoch: {} / {}'.format(epoch, args.epochs))

        for key in metrics.keys():
            metrics[key].reset()

        # training phase
        # train model
        meter_prefix = 'train'
        model = model.train()
        forward_epoch(data_loader_train, meter_prefix)
        model = model.eval()
        last_train_data_centers = evaluate_epoch(data_loader_train, meter_prefix)
        # evaluation phase
        with torch.no_grad():
            model = model.eval()
            # test set forward
            meter_prefix = 'test'
            forward_epoch(data_loader_test, meter_prefix)
            test_data_centers = evaluate_epoch(data_loader_test, meter_prefix)
            percent_improvement, early_stopping = update_early_stopping()

        model_module = model
        if not args.is_single_cuda_device:
            model_module = model.module

        # save best metrics
        if epoch == start_epoch:
            state['best_loss'] = state['test_loss']
        elif state['best_loss'] > state['test_loss']:
            state['best_loss'] = state['test_loss']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_loss.pt'))
                # # commented to save disk space
                # # save checkpoint
                # best_checkpoint['model'] = model_module.state_dict()
                # best_checkpoint['optimizer'] = optimizer_func.state_dict()
                # best_checkpoint['state'] = state
                # best_checkpoint['args'] = args
                # torch.save(best_checkpoint, os.path.join(path_run, 'best_checkpoint.pt'))

        if epoch == start_epoch:
            state['best_accuracy'] = state['test_accuracy']
            best_epoch = epoch
            best_state_dict = copy.deepcopy(model_module.state_dict())
            best_train_data_centers = last_train_data_centers
        elif state['best_accuracy'] < state['test_accuracy']:
            state['best_accuracy'] = state['test_accuracy']
            best_epoch = epoch
            best_state_dict = copy.deepcopy(model_module.state_dict())
            best_train_data_centers = last_train_data_centers
            if args.save_models:
                torch.save(best_state_dict, os.path.join(path_run, 'best_accuracy.pt'))

        if epoch == start_epoch:
            state['best_f1'] = state['test_f1']
        elif state['best_f1'] < state['test_f1']:
            state['best_f1'] = state['test_f1']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_f1.pt'))

        epoch_time = (time.time() - time_epoch) / 60.0
        percent = epoch / args.epochs
        state['epoch_time'] = epoch_time

        avg_time_epochs.append(epoch_time)
        state['avg_epoch_time'] = np.average(avg_time_epochs)
        eta = ((args.epochs - epoch) * state['avg_epoch_time'])
        time_epoch = time.time()
        state['epoch'] = epoch

        # save last model
        torch.save(model_module.state_dict(), os.path.join(path_run, f'{args.run_name}.pt'))

        # log state each epoch
        tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

        CsvUtils2.add_hparams(
            path_sequence=path_sequence,
            run_name=args.run_name,
            args_dict=args.__dict__,
            metrics_dict=state,
            global_step=epoch
        )

        logger.info(
            f'{round(percent * 100, 2)}% each: {round(state["avg_epoch_time"], 2)}min, eta: {round(eta, 2)} min, train loss: {round(state["train_loss"], 5)}, '
            f'test loss: {round(state["test_loss"], 5)}, improve: {round(percent_improvement, 3)}, train acc: {round(state["train_accuracy"], 2)}, '
            f'test acc: {round(state["test_accuracy"], 2)}, train f1: {round(state["train_f1"], 2)}, test f1: {round(state["test_f1"], 2)}')

        if early_stopping:
            break

    # run eval
    for mode in ['last', 'best']:
        train_data_centers = last_train_data_centers
        cur_epoch = epoch
        if mode == 'best':
            cur_epoch = best_epoch
            model.load_state_dict(best_state_dict)
            train_data_centers = best_train_data_centers
        eval_accuracy_self = torchmetrics.Accuracy()
        eval_confusion_meter_self = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
        eval_accuracy_train_embs = torchmetrics.Accuracy()
        eval_confusion_meter_train_embs = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
        samples = [it for it in os.listdir(args.path_validation)]
        model = model.eval()
        true_labels = []
        embeddings = []
        with torch.no_grad():
            for idx, name in enumerate(tqdm(samples, desc='Evaluate model')):
                if '.wav' not in name:
                    continue
                audio, _ = librosa.load(os.path.join(args.path_validation, name), sr=args.data_samplerate)
                if audio is None:
                    break
                else:
                    label = re.search("happiness|sadness|anger|neutral|other", name)[0]
                    if label not in args.classes:
                        continue

                    if args.gender:
                        gender = re.search("male|female", name)[0]
                        if gender != args.gender:
                            continue

                input_size = args.input_size
                step_size = int(args.input_size / 2)
                pad_reference = step_size
                if len(audio) < input_size:
                    pad_reference = input_size
                overflow = len(audio) % pad_reference
                if overflow > 0:
                    pad_len = pad_reference - overflow
                    audio = np.pad(audio, (0, pad_len))
                t_audio = torch.FloatTensor(audio).to(args.device).unsqueeze(dim=0).unsqueeze(dim=0).unsqueeze(dim=0)
                t_audio_strided = F.unfold(t_audio, kernel_size=(1, input_size), stride=(1, step_size)).squeeze().T

                if len(t_audio_strided.shape) == 1:
                    t_audio_strided = t_audio_strided.unsqueeze(dim=0)

                loader = torch.utils.data.DataLoader(
                    torch.utils.data.TensorDataset(t_audio_strided),
                    batch_size=args.batch_size,
                    shuffle=False
                )
                for batch, in loader:
                    embedding = model.forward(batch).cpu()
                    for emb in embedding:
                        embeddings.append(emb)
                        true_labels.append(data_loader_train.dataset.classes.index(label))
                        dists_to_centers = cdist(
                            np.array(list(train_data_centers.values())),
                            np.expand_dims(emb, axis=0),
                            metric='euclidean'
                        ).squeeze()
                        idx_closest = np.argmin(dists_to_centers)
                        y_prim = list(train_data_centers.keys())[idx_closest]

                        y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim,)), len(args.classes)).type(torch.FloatTensor)
                        label_t = torch.LongTensor((data_loader_train.dataset.classes.index(label),))
                        eval_accuracy_train_embs.update(y_prim_idx_each, label_t)
                        eval_confusion_meter_train_embs.update(y_prim_idx_each, label_t)

            embeddings = np.stack(embeddings)
            true_labels = np.array(true_labels)
            # get centers
            centers_by_classes = {}
            for l in set(true_labels):
                cur_l = true_labels == l
                cur_emb = embeddings[cur_l]
                # calculate initial center
                centers_by_classes[l] = np.mean(cur_emb, axis=0)

                # remove 90th percentile of samples
                distance_to_center = cdist(
                    cur_emb,
                    np.expand_dims(centers_by_classes[l], axis=0),
                    metric='euclidean'
                ).squeeze()
                indexes_sorted = np.argsort(distance_to_center)
                embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

                # recalculate centers
                centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

            # calculate accuracy
            for idx, emb in enumerate(embeddings):
                dists_to_centers = cdist(
                    np.array(list(centers_by_classes.values())),
                    np.expand_dims(emb, axis=0),
                    metric='euclidean'
                ).squeeze()
                idx_closest = np.argmin(dists_to_centers)
                y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
                # prepare data for metrics
                y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
                label = torch.LongTensor((true_labels[idx],))

                eval_accuracy_self.update(y_prim_idx_each, label)
                eval_confusion_meter_self.update(y_prim_idx_each, label)

            pca = PCA(n_components=2)
            embeddings_tb_2d = pca.fit_transform(embeddings)
            plt.cla()
            plt.clf()
            scatter = plt.scatter(embeddings_tb_2d[:, -1], embeddings_tb_2d[:, -2], s=1, c=true_labels, cmap='jet')
            plt.legend(handles=scatter.legend_elements()[0], labels=data_loader_train.dataset.classes)
            plt.tight_layout(pad=0.5)
            embedding_dir = f'{path_run}/embeddings'
            plt.savefig(f'{embedding_dir}/eval_embeddings_{mode}_self_epoch_{cur_epoch}.png')
            plt.close()

        state[f'eval_accuracy_self_{mode}'] = eval_accuracy_self.compute().item()
        state[f'eval_accuracy_train_embs_{mode}'] = eval_accuracy_train_embs.compute().item()

        tensorboard_utils.addPlotConfusionMatrix(
            dataXY=eval_confusion_meter_self.compute().numpy(),
            ticks=args.classes,
            tag=f'{mode}_eval_self_confusion_matrix',
            global_step=cur_epoch,
            is_percent_added=True
        )
        tensorboard_utils.addPlotConfusionMatrix(
            dataXY=eval_confusion_meter_train_embs.compute().numpy(),
            ticks=args.classes,
            tag=f'{mode}_eval_train_embs_confusion_matrix',
            global_step=cur_epoch,
            is_percent_added=True
        )
        logger.info(f'Evaluate {mode} results; eval_accuracy_self: {eval_accuracy_self.compute().item()},'
                    f' eval_accuracy_train_embs: {eval_accuracy_train_embs.compute().item()}')
    # log eval values at end of training
    tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

    CsvUtils2.add_hparams(
        path_sequence=path_sequence,
        run_name=args.run_name,
        args_dict=args.__dict__,
        metrics_dict=state,
        global_step=epoch
    )

    tensorboard_writer.close()
