import math
import torch
import torch.nn.functional as F
import torch.nn as nn


class OnlineTripletLoss(nn.Module):
    """
    Online Triplets loss
    Takes a batch of embeddings and corresponding labels.
    Triplets are generated using triplet_selector object that take embeddings and targets and return indices of
    triplets
    """

    def __init__(self, margin, triplet_selector, trainable_margin=False):
        super(OnlineTripletLoss, self).__init__()
        self.margin = margin
        self.is_trainable_margin = trainable_margin
        self.learnable_margin = nn.Parameter(torch.tensor(1.))
        self.triplet_selector = triplet_selector

    def get_margin(self):
        if self.is_trainable_margin:
            return torch.sigmoid(self.margin * self.learnable_margin).item()
        else:
            return self.margin

    def forward(self, embeddings, target):

        triplets = self.triplet_selector.get_triplets(embeddings, target)

        if embeddings.is_cuda:
            triplets = triplets.cuda()

        ap_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 1]]).pow(2).sum(1)  # .pow(.5)
        an_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 2]]).pow(2).sum(1)  # .pow(.5)
        margin = torch.sigmoid(self.margin * self.learnable_margin) if self.is_trainable_margin else self.margin
        losses = F.relu(ap_distances - an_distances + margin)

        ap_distances = ap_distances.cpu().detach().numpy()
        an_distances = an_distances.cpu().detach().numpy()

        return losses.mean(), ap_distances, an_distances


# https://github.com/clovaai/voxceleb_trainer
class AAMSoftmaxLoss(nn.Module):
    def __init__(self, in_features, out_features, margin=0.3, scale=15, easy_margin=False):
        super(AAMSoftmaxLoss, self).__init__()

        self.test_normalize = True

        self.m = margin
        self.s = scale
        self.in_feats = in_features
        self.weight = torch.nn.Parameter(torch.FloatTensor(out_features, in_features), requires_grad=True)
        self.ce = nn.CrossEntropyLoss()
        nn.init.xavier_normal_(self.weight, gain=1)

        self.easy_margin = easy_margin
        self.cos_m = math.cos(self.m)
        self.sin_m = math.sin(self.m)

        # make the function cos(theta+m) monotonic decreasing while theta in [0°,180°]
        self.th = math.cos(math.pi - self.m)
        self.mm = math.sin(math.pi - self.m) * self.m

        print('Initialised AAMSoftmax margin %.3f scale %.3f'%(self.m,self.s))

    def forward(self, x, label=None):
        self.to(x.device)
        assert x.size()[0] == label.size()[0]
        assert x.size()[1] == self.in_feats

        # cos(theta)
        cosine = F.linear(F.normalize(x), F.normalize(self.weight))
        # cos(theta + m)
        sine = torch.sqrt((1.0 - torch.pow(cosine, 2)).clamp(0, 1))
        phi = cosine * self.cos_m - sine * self.sin_m

        if self.easy_margin:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where((cosine - self.th) > 0, phi, cosine - self.mm)

        #one_hot = torch.zeros(cosine.size(), device='cuda' if torch.cuda.is_available() else 'cpu')
        one_hot = torch.zeros_like(cosine)
        one_hot.scatter_(1, label.view(-1, 1), 1)
        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)
        output = output * self.s

        loss = self.ce.forward(output, label)
        return loss
