import torch
import torch.nn
import torch.utils.data
import torch.utils.data.sampler
from torch.utils.data import Sampler
import numpy as np

from loguru import logger
from modules.file_utils import FileUtils


class Dataset(torch.utils.data.dataset.Dataset):

    def __init__(self, args, datasource_json):
        super().__init__()
        self.args = args
        self.classes_to_use = args.classes
        self.gender_to_use = args.gender

        self.json_desc = FileUtils.readJSON(datasource_json)

        self.mmap_x = np.memmap(
            datasource_json.replace('.json', '.mmap'),
            dtype=np.float16,
            shape=tuple(self.json_desc['shape'])
        )

        class_ids = self.json_desc['class_idxes']
        classes = self.json_desc['classes']
        genders = self.json_desc['genders']

        self.idxes_to_use = []
        self.class_ids_new = []
        self.genders = []
        self.speaker_ids = []
        self.sample_ids = []
        for idx in range(len(class_ids)):
            class_str = classes[class_ids[idx]]
            if class_str in self.classes_to_use:
                if self.gender_to_use:
                    if genders[idx] != self.gender_to_use:
                        continue
                self.idxes_to_use.append(idx)
                self.class_ids_new.append(self.classes_to_use.index(class_str))
                self.genders.append(genders[idx])
                self.speaker_ids.append(self.json_desc['speaker_ids'][idx])
                self.sample_ids.append(self.json_desc['sample_ids'][idx])

        self.idx_to_classes = list(range(len(self.classes_to_use)))
        self.class_examples = {}
        for class_id in self.idx_to_classes:
            self.class_examples[class_id] = np.where(np.array(self.class_ids_new) == class_id)[0].tolist()

        self.y_str = [self.classes_to_use[i] for i in self.class_ids_new]
        for label in set(self.y_str):
            logger.info(f'{label} has {self.y_str.count(label)} samples')

    def __len__(self):
        return len(self.idxes_to_use)

    def __getitem__(self, idx):

        y_str = self.y_str[idx]
        if y_str in self.args.classes:
            y = self.args.classes.index(y_str)
        else:
            y = self.args.classes.index('other')

        mmap_idx = self.idxes_to_use[idx]
        x = self.mmap_x[mmap_idx]  # np.float16 (min eps = 0.000977, not enough for normalization)
        x = x.astype(np.float32)  # np.float32 (min eps = 1.1920929e-07)

        x = torch.FloatTensor(x)
        y = torch.LongTensor([y])

        gender = self.genders[idx]
        speaker_id = self.speaker_ids[idx]
        sample_id = self.sample_ids[idx]

        if y_str in self.args.classes:
            class_label = y_str
        else:
            class_label = 'other'

        return x, y, gender, speaker_id, sample_id, class_label


class TripletSampler(Sampler[int]):
    def __init__(self, data_source: Dataset, shuffle: bool = False, seed: int = None):
        super().__init__(data_source)
        self.data_source = data_source
        self.shuffle = shuffle
        self.seed = seed

    def __iter__(self):
        # randomize samples
        # if test set or fixed seed shuffle to preset seed
        generator = torch.Generator()
        if self.shuffle:
            if self.seed is None:
                generator = generator.manual_seed(int(torch.empty((), dtype=torch.int64).random_().item()))
            else:
                generator = generator.manual_seed(self.seed)
        else:
            generator = generator.manual_seed(0)

        if self.shuffle:
            labels_free = torch.randperm(len(self.data_source.idx_to_classes), generator=generator).numpy().tolist()
        else:
            labels_free = list(range(len(self.data_source.idx_to_classes)))
        label_indexes = []
        for class_idx in labels_free:
            class_indexes = self.data_source.class_examples[class_idx]
            class_indexes = torch.LongTensor(list(class_indexes))
            shuffled_indexes = torch.randperm(len(class_indexes), generator=generator)
            class_indexes = torch.index_select(class_indexes, 0, shuffled_indexes).numpy().tolist()
            label_indexes.append(class_indexes)

        idx_label = 0
        while len(labels_free) >= 2:
            if len(labels_free) <= idx_label:
                idx_label = 0

            if len(label_indexes[idx_label]) < 3:
                labels_free.pop(idx_label)
                label_indexes.pop(idx_label)
            else:
                if self.shuffle:
                    samples = list(np.random.choice(label_indexes[idx_label], 3, replace=False))
                else:
                    samples = label_indexes[idx_label][:3]
                for sample in samples:
                    yield sample
                    label_indexes[idx_label].remove(sample)
            idx_label += 1

    def __len__(self) -> int:
        return len(self.data_source)


def get_data_loaders(args):
    logger.info('Using datasource_v1')

    dataset_train = Dataset(args, args.path_train_emo)
    dataset_test = Dataset(args, args.path_test_emo)

    logger.info('train dataset')
    sampler = TripletSampler(dataset_train, shuffle=True)
    data_loader_train = torch.utils.data.DataLoader(dataset_train,
                                                    batch_size=args.batch_size,
                                                    num_workers=args.data_workers,
                                                    pin_memory=True,
                                                    drop_last=True,
                                                    sampler=sampler)
    logger.info('test dataset')
    sampler = TripletSampler(dataset_test)
    data_loader_test = torch.utils.data.DataLoader(dataset_test,
                                                   batch_size=args.batch_size,
                                                   num_workers=args.data_workers,
                                                   pin_memory=True,
                                                   drop_last=True,
                                                   sampler=sampler)

    return data_loader_train, data_loader_test
