import math

import torch
from torch import nn
from torch.nn import functional as F

from modules.demucs_helpers import upsample2


def rescale_conv(conv, reference):
    std = conv.weight.std().detach()
    scale = (std / reference)**0.5
    conv.weight.data /= scale
    if conv.bias is not None:
        conv.bias.data /= scale


def rescale_module(module, reference):
    for sub in module.modules():
        if isinstance(sub, (nn.Conv1d, nn.ConvTranspose1d)):
            rescale_conv(sub, reference)


class Model(nn.Module):
    """
    Demucs speech enhancement model.
    Args:
        - chin (int): number of input channels.
        - chout (int): number of output channels.
        - hidden (int): number of initial hidden channels.
        - depth (int): number of layers.
        - kernel_size (int): kernel size for each layer.
        - stride (int): stride for each layer.
        - resample (int): amount of resampling to apply to the input/output.
            Can be one of 1, 2 or 4.
        - growth (float): number of channels is multiplied by this for every layer.
        - max_hidden (int): maximum number of channels. Can be useful to
            control the size/speed of the model.
        - normalize (bool): if true, normalize the input.
        - glu (bool): if true uses GLU instead of ReLU in 1x1 convolutions.
        - rescale (float): controls custom weight initialization.
            See https://arxiv.org/abs/1911.13254.
        - floor (float): stability flooring when normalizing.

    """
    def __init__(self, args):

        super().__init__()
        self.args = args

        if args.resample not in [1, 2, 4]:
            raise ValueError("Resample should be 1, 2 or 4.")

        self.chin = 1
        self.depth = self.args.encoder_depth
        self.hidden = self.args.encoder_hidden_size
        self.kernel_size = self.args.encoder_kernel_size
        self.stride = self.args.encoder_stride
        self.resample = self.args.resample
        self.data_scaling_type = self.args.data_scaling_type
        self.growth = self.args.encoder_growth
        self.floor = 1e-3
        self.rescale = 0.1
        self.max_hidden = 10000

        self.encoder = nn.ModuleList()
        activation = nn.GLU(1)
        ch_scale = 2

        chin = self.chin
        hidden = self.hidden
        l_out = 64_000 * self.resample
        for index in range(self.depth):
            encode = []
            encode += [
                nn.Conv1d(chin, hidden, self.kernel_size, self.stride),
                nn.ReLU(),
                nn.Conv1d(hidden, hidden * ch_scale, 1), activation,
            ]
            l_out = math.floor((l_out + 2 * 0 - 1 * (self.kernel_size - 1) - 1) / self.stride + 1)

            self.encoder.append(nn.Sequential(*encode))

            chin = hidden
            hidden = min(int(self.growth * hidden), self.max_hidden)

        if self.rescale:
            rescale_module(self, reference=self.rescale)

        linear_modules = torch.nn.ModuleList()
        linear_modules.append(torch.nn.Linear(in_features=int((l_out+1)*(hidden/2)), out_features=128))
        linear_modules.append(torch.nn.Dropout(p=0.5))
        linear_modules.append(torch.nn.BatchNorm1d(num_features=128))
        linear_modules.append(torch.nn.ReLU())
        linear_modules.append(torch.nn.Linear(in_features=128, out_features=len(self.args.classes)))
        self.linear = torch.nn.Sequential(*linear_modules)

    def valid_length(self, length):
        """
        Return the nearest valid length to use with the model so that
        there is no time steps left over in a convolutions, e.g. for all
        layers, size of the input - kernel_size % stride = 0.

        If the mixture has a valid length, the estimated sources
        will have exactly the same length.
        """
        length = math.ceil(length * self.resample)
        for idx in range(self.depth):
            length = math.ceil((length - self.kernel_size) / self.stride) + 1
            length = max(length, 1)
        for idx in range(self.depth):
            length = (length - 1) * self.stride + self.kernel_size
        length = int(math.ceil(length / self.resample))
        return int(length)

    @property
    def total_stride(self):
        return self.stride ** self.depth // self.resample

    @staticmethod
    def standardize(batch, scope='sample'):
        # scope [sample; batch]
        eps = 1e-8
        if scope == 'sample':
            for idx, sample in enumerate(batch):
                mean = torch.mean(sample)
                std = torch.std(sample)
                batch[idx] = (batch[idx] - mean) / (std + eps)
        elif scope == 'batch':
            mean = torch.mean(batch)
            std = torch.std(batch)
            batch = (batch - mean) / (std + eps)
        return batch

    @staticmethod
    def minmax_scaling(batch, scope='sample', range_min=-1, range_max=1):
        eps = 1e-8
        if scope == 'sample':
            for idx, sample in enumerate(batch):
                sample_min = torch.amin(sample)
                sample_max = torch.amax(sample)
                batch[idx] = ((batch[idx] - sample_min) / (sample_max - sample_min + eps)) * \
                             (range_max - range_min) + range_min
        elif scope == 'batch':
            batch_min = torch.amin(batch)
            batch_max = torch.amax(batch)
            batch = (batch - batch_min) / (batch_max - batch_min + eps) * (range_max - range_min) + range_min

        return batch

    def forward(self, mix):
        if self.args.data_scaling_type == 'standardization':
            mix = self.minmax_scaling(mix, scope='sample', range_min=-1, range_max=1)
        elif self.args.data_scaling_type == 'normalization':
            mix = self.standardize(x, scope='sample')

        if mix.dim() == 2:
            mix = mix.unsqueeze(1)

        length = mix.shape[-1]
        x = mix
        x = F.pad(x, (0, self.valid_length(length) - length))

        if self.resample == 2:
            x = upsample2(x)
        elif self.resample == 4:
            x = upsample2(x)
            x = upsample2(x)

        for encode in self.encoder:
            x = encode(x)

        x = torch.flatten(x, -2, -1)
        x = self.linear.forward(x)

        return x
