import torch
import torch.utils.data
import math

from scipy.signal import butter, lfilter
from torchaudio.transforms import MFCC, Spectrogram, MelSpectrogram
from modules.demucs_helpers import upsample2

class Model(torch.nn.Module):
    def __init__(self, args):
        super().__init__()

        self.args = args
        self.mean = self.args.model_mean
        self.std = self.args.model_std
        self.nobservations = 0
        self.red_dims = (0, 2, 3)
        self.resample = args.resample

        # Highpass filter audio a and b parts
        if self.args.audio_cutoff > 0:
            nyq = 0.5 * self.args.data_samplerate
            normal_cutoff = self.args.audio_cutoff / nyq
            b, a = butter(5, normal_cutoff, btype='high', analog=False)
            self.audio_a = a
            self.audio_b = b

        self.downsample = torch.nn.Upsample(size=(128, 128))
        if self.args.spectrogram_type == "spectrogram":
            self.transform = Spectrogram(n_fft=self.args.n_fft,
                                         win_length=self.args.win_length,
                                         hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "melspectrogram":
            self.transform = MelSpectrogram(sample_rate=self.args.data_samplerate,
                                            n_fft=self.args.n_fft,
                                            n_mels=self.args.n_mels,
                                            win_length=self.args.win_length,
                                            hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "mfcc":
            self.transform = MFCC(sample_rate=self.args.data_samplerate,
                                  n_mfcc=self.args.n_mfcc,
                                  melkwargs={
                                      "n_fft": self.args.n_fft,
                                      "n_mels": self.args.n_mels,
                                      "hop_length": self.args.hop_length,
                                      "win_length": self.args.win_length
                                  }).to(self.args.device)

        if not (math.log(self.args.encoder_ch_out, 2)).is_integer() or self.args.encoder_ch_out < 8:
            raise ValueError("encoder_ch_out should be a power of 2 larger or equal than 8.")

        encoder_modules = torch.nn.ModuleList()
        linear_modules = torch.nn.ModuleList()
        prev_out_channels = 1
        out_channels = 32 if self.args.encoder_ch_out >= 128 else int(self.args.encoder_ch_out/4)
        layer_count = int(math.log(self.args.encoder_ch_out//out_channels, 2) + 1)

        # ENCODER
        for layer in range(layer_count):
            encoder_modules.append(torch.nn.Conv2d(in_channels=prev_out_channels,
                                                   out_channels=out_channels,
                                                   kernel_size=3,
                                                   padding=1))
            if self.args.normalization_module != 'none':
                encoder_modules.append(self.normalization_module_2d(out_channels, self.args.normalization_module))
            encoder_modules.append(torch.nn.ReLU())

            if layer <= 1:
                encoder_modules.append(torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1))

            encoder_modules.append(torch.nn.Conv2d(in_channels=out_channels,
                                                   out_channels=out_channels,
                                                   kernel_size=3,
                                                   padding=1))
            if self.args.normalization_module != 'none':
                encoder_modules.append(self.normalization_module_2d(out_channels, self.args.normalization_module))
            encoder_modules.append(torch.nn.ReLU())

            if layer == layer_count - 1:
                encoder_modules.append(torch.nn.Conv2d(in_channels=out_channels,
                                                       out_channels=out_channels,
                                                       kernel_size=3,
                                                       padding=1))
                if self.args.normalization_module != 'none':
                    encoder_modules.append(self.normalization_module_2d(out_channels, self.args.normalization_module))
                encoder_modules.append(torch.nn.ReLU())

                encoder_modules.append(torch.nn.Conv2d(in_channels=out_channels,
                                                       out_channels=out_channels,
                                                       kernel_size=3,
                                                       padding=1))
                if self.args.normalization_module != 'none':
                    encoder_modules.append(self.normalization_module_2d(out_channels, self.args.normalization_module))
                encoder_modules.append(torch.nn.ReLU())

            if not layer == layer_count - 1:
                prev_out_channels = out_channels
                out_channels = out_channels*2

        encoder_modules.append(torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1))
        self.encoder = torch.nn.Sequential(*encoder_modules)

        linear_modules.append(torch.nn.Linear(in_features=out_channels*256, out_features=128))
        linear_modules.append(torch.nn.Dropout(p=0.5))
        linear_modules.append(torch.nn.BatchNorm1d(num_features=128))
        linear_modules.append(torch.nn.ReLU())
        linear_modules.append(torch.nn.Linear(in_features=128, out_features=len(args.classes)))
        self.linear = torch.nn.Sequential(*linear_modules)

    def forward(self, x):

        if self.args.audio_cutoff > 0:
            for idx in range(len(x)):
                x[idx] = torch.FloatTensor(lfilter(self.audio_b, self.audio_a, x[idx].cpu().numpy())).to(self.args.device)

        if self.args.data_scaling_type == 'normalization':
            x = self.minmax_scaling(x, scope='sample', range_min=-1, range_max=1)
        elif self.args.data_scaling_type == 'standardization':
            x = self.standardize(x, scope='sample')

        if self.resample == 2:
            x = upsample2(x)
        elif self.resample == 4:
            x = upsample2(x)
            x = upsample2(x)

        x_in = self.transform(x)
        # x_in = self.amplitude_to_db(x_in)
        x_in = torch.pow(torch.abs(x_in), 0.2)

        x_in = x_in.unsqueeze(dim=1)  # add channel dim
        x_in = self.downsample(x_in)
        if self.args.apply_normalization:
            x_in = self.standardize(x_in, scope=self.args.normalization_scope)

        x_out = self.encoder.forward(x_in)
        x_out = torch.flatten(x_out, -3, -1)
        x_out = self.linear.forward(x_out)

        return x_out

    def calculate_standardization_params(self, x):
        x_in = self.transform(x)
        x_in = torch.sign(x_in) * torch.pow(torch.abs(x_in), 0.2)
        x_in = x_in.unsqueeze(dim=1)  # add channel dim
        x_in = self.downsample(x_in)

        if self.nobservations == 0:
            self.mean = torch.mean(x_in, dim=self.red_dims, keepdim=True)
            self.std = torch.std(x_in, dim=self.red_dims, keepdim=True)
            self.nobservations = x_in.shape[0]
            self.ndimensions = x_in.shape[1]
        else:

            # find mean of new mini batch
            newmean = torch.mean(x_in, dim=self.red_dims, keepdim=True)
            newstd = torch.std(x_in, dim=self.red_dims, keepdim=True)

            # update number of observations
            m = self.nobservations
            n = x_in.shape[0]

            # update running statistics
            tmp = self.mean
            self.mean = m / (m + n) * tmp + n / (m + n) * newmean
            self.std = m / (m + n) * self.std ** 2 + n / (m + n) * newstd ** 2 + \
                       m * n / (m + n) ** 2 * (tmp - newmean) ** 2
            self.std = torch.sqrt(self.std)

            # update total number of seen samples
            self.nobservations += n

    def standardize(self, batch, scope='sample'):
        # scope [sample; batch; dataset]
        eps = 1e-8
        if scope == 'sample':
            for idx, sample in enumerate(batch):
                mean = torch.mean(sample)
                std = torch.std(sample)
                batch[idx] = (batch[idx] - mean) / (std + eps)
        elif scope == 'batch':
            mean = torch.mean(batch)
            std = torch.std(batch)
            batch = (batch - mean) / (std + eps)
        elif scope == 'dataset':
            batch = (batch - self.mean) / (self.std + eps)
        return batch

    @staticmethod
    def minmax_scaling(batch, scope='sample', range_min=-1, range_max=1):
        eps = 1e-8
        if scope == 'sample':
            for idx, sample in enumerate(batch):
                sample_min = torch.amin(sample)
                sample_max = torch.amax(sample)
                batch[idx] = ((batch[idx] - sample_min) / (sample_max - sample_min + eps)) * \
                             (range_max - range_min) + range_min
        elif scope == 'batch':
            batch_min = torch.amin(batch)
            batch_max = torch.amax(batch)
            batch = (batch - batch_min) / (batch_max - batch_min + eps) * (range_max - range_min) + range_min

        return batch

    @staticmethod
    def normalization_module_2d(out_channels, type='batch'):
        if type == 'batch':
            return torch.nn.BatchNorm2d(num_features=out_channels)
        elif type == 'instance':
            return torch.nn.InstanceNorm2d(num_features=out_channels)

    @staticmethod
    def normalization_module_1d(out_channels, type='batchnorm'):
        if type == 'batch':
            return torch.nn.BatchNorm1d(num_features=out_channels)
        elif type == 'instance':
            return torch.nn.InstanceNorm1d(num_features=16, affine=True)
