from torchvision.ops import StochasticDepth
from torch.nn import functional as nnF
import torch.utils.data
from torchaudio.transforms import MFCC, Spectrogram, MelSpectrogram
from scipy.signal import butter, lfilter

from modules.audio_augmentations import *


class VectorQuantizer(nn.Module):
    """
    Discretization bottleneck part of the VQ-VAE.
    Inputs:
    - n_e : number of embeddings
    - e_dim : dimension of embedding
    - beta : commitment cost used in loss term, beta * ||z_e(x)-sg[e]||^2
    """

    def __init__(self, n_e, e_dim, beta, args):
        super().__init__()
        self.n_e = n_e
        self.e_dim = e_dim
        self.beta = beta
        self.args = args

        self.embedding = nn.Embedding(self.n_e, self.e_dim)
        self.embedding.weight.data.uniform_(-1.0 / self.n_e, 1.0 / self.n_e)

    def forward(self, z):
        """
        Inputs the output of the encoder network z and maps it to a discrete
        one-hot vector that is the index of the closest embedding vector e_j
        z (continuous) -> z_q (discrete)
        z.shape = (batch, channel, height, width)
        quantization pipeline:
            1. get encoder input (B,C,H,W)
            2. flatten input to (B*H*W,C)
        """

        """ reshape z -> (batch, height, width, channel) and flatten """
        z = z.permute(0, 2, 3, 1).contiguous()
        z_flattened = z.view(-1, self.e_dim)

        """ distances from z to embeddings e_j (z - e)^2 = z^2 + e^2 - 2 e * z """
        d = torch.sum(z_flattened ** 2, dim=1, keepdim=True) + \
            torch.sum(self.embedding.weight**2, dim=1) - 2 * \
            torch.matmul(z_flattened, self.embedding.weight.t())

        """ find closest encodings """
        min_encoding_indices = torch.argmin(d, dim=1).unsqueeze(1)
        min_encodings = torch.zeros(min_encoding_indices.shape[0], self.n_e).to(self.args.device)
        min_encodings.scatter_(1, min_encoding_indices, 1)

        """ get quantized latent vectors """
        z_q = torch.matmul(min_encodings, self.embedding.weight).view(z.shape)

        """ compute loss for embedding """
        loss = torch.mean((z_q.detach() - z) ** 2) + self.beta * torch.mean((z_q - z.detach()) ** 2)

        """ preserve gradients """
        z_q = z + (z_q - z).detach()

        """ perplexity """
        e_mean = torch.mean(min_encodings, dim=0)
        perplexity = torch.exp(-torch.sum(e_mean * torch.log(e_mean + 1e-10)))

        """ reshape back to match original input shape """
        z_q = z_q.permute(0, 3, 1, 2).contiguous()

        return loss, z_q, perplexity, min_encodings, min_encoding_indices


class Model(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = VGG_Encoder()
        self.decoder = Decoder(args)
        self.quantize_conv = nn.Conv2d(args.encoder_hidden, args.embedding_dim, kernel_size=1)
        self.conv_reduce = nn.Conv2d(args.embedding_dim, args.embedding_dim, kernel_size=4)
        self.vector_quantization = VectorQuantizer(args.num_embeddings, args.embedding_dim, args.beta, args)

        # Highpass filter audio a and b parts
        if self.args.audio_cutoff > 0:
            nyq = 0.5 * self.args.data_samplerate
            normal_cutoff = self.args.audio_cutoff / nyq
            b, a = butter(5, normal_cutoff, btype='high', analog=False)
            self.audio_a = a
            self.audio_b = b

        self.downsample = torch.nn.Upsample(size=(128, 128))
        if self.args.spectrogram_type == "spectrogram":
            self.transform = Spectrogram(n_fft=self.args.n_fft,
                                         win_length=self.args.win_length,
                                         hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "melspectrogram":
            self.transform = MelSpectrogram(sample_rate=self.args.data_samplerate,
                                            n_fft=self.args.n_fft,
                                            n_mels=self.args.n_mels,
                                            win_length=self.args.win_length,
                                            hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "mfcc":
            self.transform = MFCC(sample_rate=self.args.data_samplerate,
                                  n_mfcc=self.args.n_mfcc,
                                  melkwargs={
                                      "n_fft": self.args.n_fft,
                                      "n_mels": self.args.n_mels,
                                      "hop_length": self.args.hop_length,
                                      "win_length": self.args.win_length
                                  }).to(self.args.device)

    def standardize(self, batch):
        eps = 1e-8
        for idx, sample in enumerate(batch):
            mean = torch.mean(sample)
            std = torch.std(sample)
            batch[idx] = (batch[idx] - mean) / (std + eps)
        return batch

    @staticmethod
    def minmax_scaling(batch, range_min=-1, range_max=1):
        eps = 1e-8
        for idx, sample in enumerate(batch):
            sample_min = torch.amin(sample)
            sample_max = torch.amax(sample)
            batch[idx] = ((batch[idx] - sample_min) / (sample_max - sample_min + eps)) * \
                         (range_max - range_min) + range_min
        return batch

    def forward(self, x):
        if self.args.audio_cutoff > 0:
            for idx in range(len(x)):
                x[idx] = torch.FloatTensor(lfilter(self.audio_b, self.audio_a, x[idx].cpu().numpy()))
        x = x.to(self.args.device)
        x = self.standardize(x)

        x_in = self.transform(x)
        x_in = torch.pow(torch.abs(x_in), 0.2)

        x_in = x_in.unsqueeze(dim=1)  # add channel dim
        x_in = self.downsample(x_in)
        x_in = self.minmax_scaling(x_in, range_min=-1, range_max=1)

        z_e = self.encoder.forward(x_in)
        z_e = self.quantize_conv.forward(z_e)
        embedding_loss, z_q, perplexity, _, _ = self.vector_quantization.forward(z_e)
        z_q_out = self.conv_reduce.forward(z_q).squeeze()
        z_q = z_q.contiguous().view(z_q.size(0), -1)

        x_dec = self.decoder.forward(z_q)

        return embedding_loss, x_dec, x_in, perplexity, z_q_out


VGG_types = {
    "VGG11": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG13": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG16": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        1024,
        1024,
        1024,
        "M",
    ],
    "VGG19": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


class VGG_Encoder(nn.Module):
    def __init__(self, in_channels=1):
        super().__init__()
        self.in_channels = in_channels
        self.conv_layers = self.create_conv_layers(VGG_types["VGG16"])

    def forward(self, x):
        return self.conv_layers.forward(x)

    def create_conv_layers(self, architecture):
        layers = []
        in_channels = self.in_channels

        for x in architecture:
            if type(x) == int:
                out_channels = x

                layers += [
                    nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
                    nn.BatchNorm2d(num_features=out_channels),
                    nn.ELU()
                ]
                in_channels = x
            elif x == "M":
                layers += [nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2))]

        return nn.Sequential(*layers)


class Decoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.decoder_lin = nn.Linear(in_features=args.embedding_dim * 4 * 4, out_features=args.encoder_hidden * 4 * 4)
        self.unflatten = nn.Unflatten(dim=1, unflattened_size=(args.encoder_hidden, 4, 4))
        self.dropout = nn.Dropout(p=0.25)
        self.decoder_conv = nn.Sequential(
            nn.ConvTranspose2d(in_channels=args.encoder_hidden, out_channels=512, kernel_size=4, stride=2, padding=1),
            nn.BatchNorm2d(num_features=512),
            nn.ELU(),
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=512),
            nn.ELU(),

            nn.ConvTranspose2d(in_channels=512, out_channels=256, kernel_size=4, stride=2, padding=1),
            nn.BatchNorm2d(num_features=256),
            nn.ELU(),
            nn.Conv2d(in_channels=256, out_channels=256, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=256),
            nn.ELU(),

            nn.ConvTranspose2d(in_channels=256, out_channels=128, kernel_size=4, stride=2, padding=1),
            nn.BatchNorm2d(num_features=128),
            nn.ELU(),
            nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=128),
            nn.ELU(),

            nn.ConvTranspose2d(in_channels=128, out_channels=64, kernel_size=4, stride=2, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ELU(),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ELU(),

            nn.ConvTranspose2d(in_channels=64, out_channels=1, kernel_size=4, stride=2, padding=1),
            nn.Tanh()
        )

    def forward(self, x):
        x_lin = self.decoder_lin.forward(x)
        x_lin = self.dropout.forward(x_lin)
        x_unflat = self.unflatten.forward(x_lin)
        x_dec = self.decoder_conv.forward(x_unflat)
        return x_dec

