import math
import torch
from torch import nn
from torch.nn import functional as F
import numpy as np
from torchaudio.transforms import MFCC, Spectrogram, MelSpectrogram, AmplitudeToDB
import torchvision


def get_param_count(model):
    params = list(model.parameters())
    result = 0
    for param in params:
        count_param = np.prod(param.size())
        result += count_param
    return result

def sinc(t):
    """sinc.

    :param t: the input tensor
    """
    return torch.where(t == 0, torch.tensor(1., device=t.device, dtype=t.dtype), torch.sin(t) / t)


def kernel_upsample2(zeros=56):
    """kernel_upsample2.

    """
    win = torch.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = torch.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t *= math.pi
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def upsample2(x, zeros=56):
    """
    Upsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    *other, time = x.shape
    kernel = kernel_upsample2(zeros).to(x)
    out = F.conv1d(x.view(-1, 1, time), kernel, padding=zeros)[..., 1:].view(*other, time)
    y = torch.stack([x, out], dim=-1)
    return y.view(*other, -1)


def kernel_downsample2(zeros=56):
    """kernel_downsample2.

    """
    win = torch.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = torch.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t.mul_(math.pi)
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def downsample2(x, zeros=56):
    """
    Downsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    if x.shape[-1] % 2 != 0:
        x = F.pad(x, (0, 1))
    xeven = x[..., ::2]
    xodd = x[..., 1::2]
    *other, time = xodd.shape
    kernel = kernel_downsample2(zeros).to(x)
    out = xeven + F.conv1d(xodd.view(-1, 1, time), kernel, padding=zeros)[..., :-1].view(
        *other, time)
    return out.view(*other, -1).mul(0.5)


def rescale_conv(conv, reference):
    std = conv.weight.std().detach()
    scale = (std / reference)**0.5
    conv.weight.data /= scale
    if conv.bias is not None:
        conv.bias.data /= scale


def rescale_module(module, reference):
    for sub in module.modules():
        if isinstance(sub, (nn.Conv1d, nn.ConvTranspose1d)):
            rescale_conv(sub, reference)

def torch_normalization(raw_x,range):
    min_x = torch.min(raw_x)
    max_x = torch.max(raw_x)
    normalized_x = range[0] + ((raw_x - min_x)*(range[1]-range[0])) / (max_x - min_x + 1e-8)
    return normalized_x

def torch_standardization(raw_x):
    mean = torch.mean(raw_x)
    std = torch.std(raw_x)
    standardized_x = ( raw_x - mean ) / std
    return standardized_x

class Model(nn.Module):
    def __init__(self, args):
        super().__init__()

        if not (math.log(args.resample,2)).is_integer(): #check if resample amount is a power of 2
            raise ValueError("Resample should be a power of 2, such as 2,4,8,16 etc.")

        self.args = args
        self.n_fft = self.args.n_fft
        self.n_mels = self.args.n_mels
        self.n_mfcc = self.args.n_mfcc
        self.resample = self.args.resample
        self.device = self.args.device
        self.z_dim = self.args.z_dim
        self.n_observations = 0
        self.mean = 0
        self.std = 0

        if self.args.model_type == 'resnext':
            self.model = torchvision.models.resnext101_32x8d(pretrained=self.args.is_resnet_pretrained)

            self.model.fc = torch.nn.Sequential(
                torch.nn.Linear(in_features=2048, out_features=2048),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.BatchNorm1d(num_features=2048),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=2048, out_features=self.z_dim)
            )

        else:
            self.model = torchvision.models.resnet18(pretrained=self.args.is_resnet_pretrained)

            self.model.fc = torch.nn.Sequential(
                torch.nn.Linear(in_features=512, out_features=512),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.BatchNorm1d(num_features=512),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=512, out_features=args.z_dim)
            )

        self.amplitude_to_db = AmplitudeToDB()

        self.melspec = MelSpectrogram(sample_rate=16_000,
                 n_fft=self.n_fft,
                 n_mels=self.n_mels
        )

        self.mfcc = MFCC(
            sample_rate=16_000,
            n_mfcc=self.n_mfcc
        )

        self.normalize = torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                                            std=[0.229, 0.224, 0.225])


    def global_standardization(self, x):

        if self.args.is_mel:
            x = self.melspec(x)
            x = self.amplitude_to_db(x)
        else:
            x = self.mfcc(x)
            x = self.amplitude_to_db(x)

        if self.n_observations == 0:
            self.mean = torch.mean(x)
            self.std = torch.std(x)
            self.n_observations = x.shape[0]
        else:
            # find mean of new mini batch
            new_mean = torch.mean(x)
            new_std = torch.std(x)

            # update number of observations
            m = self.n_observations
            n = x.shape[0]

            # update running statistics
            tmp = self.mean
            self.mean = m / (m + n) * tmp + n / (m + n) * new_mean
            self.std = m / (m + n) * self.std ** 2 + n / (m + n) * new_std ** 2 + \
                       m * n / (m + n) ** 2 * (tmp - new_mean) ** 2
            self.std = torch.sqrt(self.std)

            # update total number of seen samples
            self.n_observations += n


    def forward_stft(self, x):

        melspectrogram = self.mfcc(x)
        melspectrogram = self.amplitude_to_db(melspectrogram)

        if self.args.standardization == 'batch_standardization':
            melspectrogram = torch_standardization(melspectrogram)
        elif self.args.standardization == 'global_standardization' :
            melspectrogram = (melspectrogram - self.mean) / self.std
        else:
            melspectrogram = torch_normalization(melspectrogram, range=(-1,1))

        return melspectrogram

    def forward(self, mix):

        for k in range(int(math.log(self.resample, 2))):
            mix = downsample2(mix)

        mix = mix.squeeze(1)
        x_spectra = self.forward_stft(mix)

        x_spectra = x_spectra.unsqueeze(1)
        x_spectra = x_spectra.expand((x_spectra.shape[0], 3, x_spectra.shape[-2], x_spectra.shape[-1]))

        if self.args.is_resnet_pretrained:
            x_spectra = torch_normalization(x_spectra, range=(0,1))
            x_spectra = self.normalize(x_spectra)

        output = self.model.forward(x_spectra)

        return output

