import math
import torch
from torch import nn
from torch.nn import functional as F
import numpy as np
from transformers import AutoModel, Wav2Vec2Model, Wav2Vec2ForSequenceClassification, Wav2Vec2Config #pip install transformers

def get_param_count(model):
    params = list(model.parameters())
    result = 0
    for param in params:
        count_param = np.prod(param.size())
        result += count_param
    return result

def sinc(t):
    """sinc.

    :param t: the input tensor
    """
    return torch.where(t == 0, torch.tensor(1., device=t.device, dtype=t.dtype), torch.sin(t) / t)


def kernel_downsample2(zeros=56):
    """kernel_downsample2.

    """
    win = torch.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = torch.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t.mul_(math.pi)
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def downsample2(x, zeros=56):
    """
    Downsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    if x.shape[-1] % 2 != 0:
        x = F.pad(x, (0, 1))
    xeven = x[..., ::2]
    xodd = x[..., 1::2]
    *other, time = xodd.shape
    kernel = kernel_downsample2(zeros).to(x)
    out = xeven + F.conv1d(xodd.view(-1, 1, time), kernel, padding=zeros)[..., :-1].view(
        *other, time)
    return out.view(*other, -1).mul(0.5)

def torch_normalization(raw_x,range):
    min_x = torch.min(raw_x)
    max_x = torch.max(raw_x)
    normalized_x = range[0] + ((raw_x - min_x)*(range[1]-range[0])) / (max_x - min_x + 1e-8)
    return normalized_x


class Model(nn.Module):
    def __init__(self, args):
        super().__init__()

        if not (math.log(args.resample,2)).is_integer(): #check if resample amount is a power of 2
            raise ValueError("Resample should be a power of 2, such as 2,4,8,16 etc.")

        self.args = args
        self.resample = self.args.resample
        self.device = self.args.device
        self.z_dim = self.args.z_dim
        self.frozen_layers = self.args.frozen_layers
        self.training_type = self.args.training_type
        self.n_observations = 0
        self.mean = 0
        self.std = 0
        self.which_finetuned_model = self.args.which_finetuned_model

        if self.which_finetuned_model == 'russian':
            # the russian finetuned model
            self.model = AutoModel.from_pretrained("Aniemore/wav2vec2-xlsr-53-russian-emotion-recognition", trust_remote_code=True)

            # replacing the original fc layer
            self.model.classifier = torch.nn.Sequential(
                torch.nn.Linear(in_features=1024, out_features=1024),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.LayerNorm(normalized_shape=1024),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=1024, out_features=self.z_dim)
            )

        elif self.which_finetuned_model == 'xlsr':
            # base multi-lingual model, PRE-TRAINED, but no finetuning
            self.model = Wav2Vec2ForSequenceClassification.from_pretrained("facebook/wav2vec2-large-xlsr-53")
            self.model.projector = torch.nn.Sequential(
                torch.nn.Linear(in_features=1024, out_features=1024),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.LayerNorm(normalized_shape=1024),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=1024, out_features=self.z_dim)
            )
            self.model.classifier = torch.nn.Identity()

        elif self.which_finetuned_model == 'xlsr_base_big':
            # non pre-trained, 316M params
            self.configuration = Wav2Vec2Config.from_pretrained("facebook/wav2vec2-large-xlsr-53")
            self.model = Wav2Vec2Model(self.configuration)
            self.classifier = torch.nn.Sequential(
                torch.nn.Linear(in_features=1024, out_features=1024),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.LayerNorm(normalized_shape=1024),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=1024, out_features=self.z_dim)
            )

        elif self.which_finetuned_model == 'wav2vec_base_small':
            # non pre-trained, 94M params
            self.configuration = Wav2Vec2Config()
            self.model = Wav2Vec2Model(self.configuration)
            self.classifier = torch.nn.Sequential(
                torch.nn.Linear(in_features=768, out_features=768),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.LayerNorm(normalized_shape=768),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=768, out_features=self.z_dim)
            )

        else:
            # use the audeering model
            self.model = Wav2Vec2Model.from_pretrained('audeering/wav2vec2-large-robust-12-ft-emotion-msp-dim')

            # audeering model is loaded without an fc layer, has to be added
            self.classifier = torch.nn.Sequential(
                torch.nn.Linear(in_features=1024, out_features=1024),
                torch.nn.Dropout(p=self.args.dropout),
                torch.nn.LayerNorm(normalized_shape=1024),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=1024, out_features=self.z_dim)
            )


        if self.frozen_layers:
            # freeze all except last fc
            for param in self.model.parameters():
                param.requires_grad = False

            if self.which_finetuned_model == 'russian':
                for param in self.model.classifier.parameters():
                    param.requires_grad = True
        else:
            # freeze only for pre-trained models
            if self.which_finetuned_model not in ['wav2vec_base_small', 'xlsr_base_big']:
                # freeze only the conv layers - the best way according to paper https://arxiv.org/pdf/2111.02735.pdf
                if self.which_finetuned_model == 'russian' or self.which_finetuned_model == 'xlsr':
                    layers = self.model.wav2vec2.feature_extractor.parameters()
                else:
                    layers = self.model.feature_extractor.conv_layers.parameters()

                for param in layers:
                    param.requires_grad = False


    def forward(self, mix):

        if not mix.max() == 1 or not mix.min == -1:
            mix = torch_normalization(mix, range=(-1, 1))

        for k in range(int(math.log(self.resample, 2))):
            mix = downsample2(mix)

        mix = mix.squeeze(1)

        std = torch.std(mix, dim=-1, keepdim=True)
        mean = torch.mean(mix, dim=-1, keepdim=True)
        standardized_audio = ((mix - mean) / (std + 1e-8))

        if self.which_finetuned_model == 'russian' or self.which_finetuned_model == 'xlsr':
            logits = self.model(standardized_audio).logits
        else:
            # audeering, wav2vec_base_small, xlsr_base_big
            output = self.model(standardized_audio)

            hidden_states = output[0]
            hidden_states = torch.mean(hidden_states, dim=1)
            logits = self.classifier(hidden_states)

        return logits
